'''
Module holding CVPredict class
'''
from typing import Optional

import pandas as pnd
import numpy
import tqdm

from ROOT import RDataFrame

import dmu.ml.utilities     as ut
import dmu.ml.cv_classifier as CVClassifier

from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:ml:cv_predict')
# ---------------------------------------
class CVPredict:
    '''
    Class used to get classification probabilities from ROOT
    dataframe and a set of models. The models were trained with CVClassifier
    '''
    def __init__(self, models : Optional[list] = None, rdf : Optional[RDataFrame] = None):
        '''
        Will take a list of CVClassifier models and a ROOT dataframe
        '''

        if models is None:
            raise ValueError('No list of models passed')

        if rdf is None:
            raise ValueError('No ROOT dataframe passed')

        self._l_model = models
        self._rdf     = rdf

        self._arr_patch : numpy.ndarray
    # --------------------------------------------
    def _get_df(self):
        '''
        Will make ROOT rdf into dataframe and return it
        '''
        model = self._l_model[0]
        l_ft  = model.features
        d_data= self._rdf.AsNumpy(l_ft)
        df_ft = pnd.DataFrame(d_data)
        df_ft = ut.patch_and_tag(df_ft)

        if 'patched_indices' in df_ft.attrs:
            self._arr_patch = df_ft.attrs['patched_indices']

        nfeat = len(l_ft)
        log.info(f'Found {nfeat} features')
        for name in l_ft:
            log.debug(name)

        return df_ft
    # --------------------------------------------
    def _non_overlapping_hashes(self, model, df_ft):
        '''
        Will return True if hashes of model and data do not overlap
        '''

        s_mod_hash = model.hashes
        s_dff_hash = ut.get_hashes(df_ft)

        s_int = s_mod_hash.intersection(s_dff_hash)
        if len(s_int) == 0:
            return True

        return False
    # --------------------------------------------
    def _predict_with_overlap(self, df_ft : pnd.DataFrame) -> numpy.ndarray:
        '''
        Takes pandas dataframe with features

        Will return numpy array of prediction probabilities when there is an overlap
        of data and model hashes
        '''
        df_ft      = ut.index_with_hashes(df_ft)
        d_prob     = {}
        ntotal     = len(df_ft)
        log.debug(30 * '-')
        log.info(f'Total size: {ntotal}')
        log.debug(30 * '-')
        for model in tqdm.tqdm(self._l_model, ascii=' -'):
            d_prob_tmp = self._evaluate_model(model, df_ft)
            d_prob.update(d_prob_tmp)

        ndata  = len(df_ft)
        nprob  = len(d_prob)
        if ndata != nprob:
            log.warning(f'Dataset size ({ndata}) and probabilities size ({nprob}) differ, likely there are repeated entries')

        l_prob = [ d_prob[hsh] for hsh in df_ft.index ]

        return numpy.array(l_prob)
    # --------------------------------------------
    def _evaluate_model(self, model : CVClassifier, df_ft : pnd.DataFrame) -> dict[str, float]:
        '''
        Evaluate the dataset for one of the folds, by taking the model and the full dataset
        '''
        s_dat_hash = set(df_ft.index)
        s_mod_hash = model.hashes

        s_dif_hash = s_dat_hash - s_mod_hash

        ndif = len(s_dif_hash)
        ndat = len(s_dat_hash)
        nmod = len(s_mod_hash)
        log.debug(f'{ndif:<20}{"=":10}{ndat:<20}{"-":10}{nmod:<20}')

        df_ft_group= df_ft.loc[df_ft.index.isin(s_dif_hash)]

        l_prob = model.predict_proba(df_ft_group)
        l_hash = list(df_ft_group.index)
        d_prob = dict(zip(l_hash, l_prob))
        nfeat  = len(df_ft_group)
        nprob  = len(l_prob)
        log.debug(f'{nfeat:<10}{"->":10}{nprob:<10}')

        return d_prob
    # --------------------------------------------
    def _patch_probabilities(self, arr_prb : numpy.ndarray) -> numpy.ndarray:
        if not hasattr(self, '_arr_patch'):
            return arr_prb

        nentries = len(self._arr_patch)
        log.warning(f'Patching {nentries} probabilities')
        arr_prb[self._arr_patch] = -1

        return arr_prb
    # --------------------------------------------
    def predict(self) -> numpy.ndarray:
        '''
        Will return array of prediction probabilities for the signal category
        '''
        df_ft = self._get_df()
        model = self._l_model[0]

        if self._non_overlapping_hashes(model, df_ft):
            log.debug('No intersecting hashes found between model and data')
            arr_prb = model.predict_proba(df_ft)
        else:
            log.info('Intersecting hashes found between model and data')
            arr_prb = self._predict_with_overlap(df_ft)

        arr_prb = self._patch_probabilities(arr_prb)

        return arr_prb
# ---------------------------------------



'''
Script used to compare ROOT files
'''

import re
import os
from dataclasses import dataclass
from typing import ClassVar

import argparse

import yaml
import numpy
from dmu.logging.log_store import LogStore

from ROOT import TFile, TTree, RDataFrame

import dmu.rfile.utilities   as rfut


log=LogStore.add_logger('rx_scripts:compare_files')
#------------------
@dataclass
class Data:
    '''
    Class used to store shared attributes
    '''
    max_entries   : int
    max_trees     : int
    l_exclude     : list[str]
    raise_if_diff : bool
    file_name_1   : str
    file_name_2   : str

    d_summary     : ClassVar[dict]= {}
#------------------
def _print_trees_difference(l_val_1 : list[str], l_val_2 : list[str]) -> None:
    s_val_1 = set(l_val_1)
    s_val_2 = set(l_val_2)

    s_only_1 = s_val_1 - s_val_2
    s_only_2 = s_val_2 - s_val_1

    Data.d_summary[f'Trees only in {Data.file_name_1}'] = list(s_only_1)
    Data.d_summary[f'Trees only in {Data.file_name_2}'] = list(s_only_2)

    nonly_1  = len(s_only_1)
    nonly_2  = len(s_only_2)

    if nonly_1 > 0:
        log.info(f'Found {nonly_1} trees in first file but not second')
        for name in s_only_1:
            log.info(f'{"":<4}{name:<20}')

    if nonly_2 > 0:
        log.info(f'Found {nonly_2} trees in second file but not first')
        for name in s_only_2:
            log.info(f'{"":<4}{name:<20}')
#------------------
def _check_trees(d_tree_1 : dict[str, TTree], d_tree_2 : dict[str, TTree]):
    '''
    Check if dictionaries have same trees
    For corresponding trees, check if number of entries is the same
    '''
    l_treename_1 = list(d_tree_1.keys())
    l_treename_2 = list(d_tree_2.keys())

    if l_treename_1 != l_treename_2:
        log.warning('Files contain different trees')
        _print_trees_difference(l_treename_1, l_treename_2)

    s_treename_1 = set(l_treename_1)
    s_treename_2 = set(l_treename_2)
    s_treename   = s_treename_1 & s_treename_2

    for treename in s_treename:
        if treename in Data.l_exclude:
            continue

        tree_1 = d_tree_1[treename]
        tree_2 = d_tree_2[treename]

        entries_1 = tree_1.GetEntries()
        entries_2 = tree_2.GetEntries()

        if entries_1 != entries_2:
            raise ValueError(f'Tree {treename} differs in entries {entries_1}/{entries_2}')

    return list(s_treename)
#------------------
def _get_data(tree : TTree) -> dict[str, numpy.ndarray]:
    rdf = RDataFrame(tree)
    if Data.max_entries > 0:
        log.warning(f'Limiting to {Data.max_entries} entries')
        rdf = rdf.Range(Data.max_entries)

    d_data = rdf.AsNumpy(exclude=[])

    return d_data
#------------------
def _check_branches(tree_name : str, l_branch_1 : list[str], l_branch_2 : list[str]) -> None:
    '''
    Takes lists of branch names
    Checks if they are the same, if not print differences

    if raise_if_diff is True, will raise exception if branches are not the same
    '''
    if l_branch_1 == l_branch_2:
        return

    s_branch_1 = set(l_branch_1)
    s_branch_2 = set(l_branch_2)

    s_branch_1_m_2 = s_branch_1.difference(s_branch_2)
    log.info(f'Found len({s_branch_1_m_2}) branches in first tree but not second')
    for branch_name in s_branch_1_m_2:
        log.debug(f'{"":<4}{branch_name:<20}')

    s_branch_2_m_1 = s_branch_2.difference(s_branch_1)
    log.info(f'Found len({s_branch_2_m_1}) branches in second tree but not first')
    for branch_name in s_branch_2_m_1:
        log.debug(f'{"":<4}{branch_name:<20}')

    Data.d_summary[tree_name] = {
            f'Only {Data.file_name_1}' : list(s_branch_1_m_2),
            f'Only {Data.file_name_2}' : list(s_branch_2_m_1),
            }

    if Data.raise_if_diff:
        raise ValueError('Branches differ')
#------------------
def _compare_branches(tree_name : str, d_data_1 : dict[str, list], d_data_2 : dict[str, list]) -> list[str]:
    '''
    Will check for different branches in trees
    Will return list of branch names for common branches
    '''
    l_branch_1 = list(d_data_1.keys())
    l_branch_2 = list(d_data_2.keys())

    l_branch_1.sort()
    l_branch_2.sort()
    _check_branches(tree_name, l_branch_1, l_branch_2)

    s_branch_1 = set(l_branch_1)
    s_branch_2 = set(l_branch_2)

    s_branch = s_branch_1.intersection(s_branch_2)

    return list(s_branch)
#------------------
def _compare(tree_name : str, d_data_1, d_data_2) -> None:
    log.info('')
    log.debug('Comparing branches')
    l_branch_name = _compare_branches(tree_name, d_data_1, d_data_2)

    log.debug('Comparing contents of branches')
    l_diff_branch = []
    for branch_name in l_branch_name:
        arr_val_1 = d_data_1[branch_name]
        arr_val_2 = d_data_2[branch_name]

        if _contents_differ(tree_name, branch_name, arr_val_1, arr_val_2):
            l_diff_branch.append(branch_name)

    ndiff = len(l_diff_branch)
    ntot  = len(l_branch_name)

    Data.d_summary[f'Branches that differ for tree: {tree_name}'] = l_diff_branch

    if ndiff == 0:
        log.debug(f'Trees {tree_name} have same contents')
        return

    log.warning(f'{ndiff:<10}{"differing branches out of":<20}{ntot:<10}{"in":<10}{tree_name:<50}')
    for branch_name in l_diff_branch:
        log.debug(f'{"":<4}{branch_name:<20}')
#------------------
def _contents_differ(tree_name : str, branch_name : str, arr_val_1 : numpy.ndarray, arr_val_2 : numpy.ndarray) -> bool:
    is_different = False
    str_type = str(arr_val_1.dtype)
    if   str_type == 'object':
        return is_different

    if str_type not in ['bool', 'int32', 'uint32', 'uint64', 'float64', 'float32']:
        log.info(f'Skipping {branch_name}, {str_type}')
        return is_different

    if not numpy.array_equal(arr_val_1, arr_val_2):
        is_different = True

        log.debug(20 * '-')
        log.debug(f'Branch {branch_name} in tree {tree_name} differ')
        log.debug(20 * '-')
        log.debug(arr_val_1)
        log.debug(arr_val_2)
        log.debug(20 * '-')

    return is_different
#------------------
def _update_keys(d_tree):
    d_out = {}

    for key, val in d_tree.items():
        #Remove everything before .root/ and use it as new key
        new_key = re.sub(r'^.*\.root/', '', key)
        d_out[new_key] = val

    return d_out
#------------------
def _check_file_existence(path : str) -> None:
    if not os.path.isfile(path):
        raise FileNotFoundError(f'Cannot find {path}')
#------------------
def _validate(file_1 : str, file_2 : str) -> None:
    _check_file_existence(file_1)
    _check_file_existence(file_2)

    ifile_1     = TFile(file_1)
    ifile_2     = TFile(file_2)

    d_tree_1    =  rfut.get_trees_from_file(ifile_1)
    d_tree_1    = _update_keys(d_tree_1)

    d_tree_2    =  rfut.get_trees_from_file(ifile_2)
    d_tree_2    = _update_keys(d_tree_2)

    l_tree_name = _check_trees(d_tree_1, d_tree_2)

    if Data.max_trees > -1:
        log.warning(f'Limiting to {Data.max_trees} trees')
        l_tree_name = l_tree_name[:Data.max_trees]

    ncommon = len(l_tree_name)
    log.debug(f'Found common {ncommon} trees')
    for name in l_tree_name:
        log.debug(f'{"":<4}{name}')

    log.info('Checking trees')
    for treename in l_tree_name:
        if treename in Data.l_exclude:
            log.debug(f'Skipping {treename}')
            continue

        log.debug(f'{"":<4}{treename}')

        tree_1 = d_tree_1[treename]
        tree_2 = d_tree_2[treename]

        log.debug('Getting data from reference')
        d_data_1= _get_data(tree_1)

        log.debug('Getting data from new')
        d_data_2= _get_data(tree_2)

        log.debug(f'Comparing {treename}')
        _compare(treename, d_data_1, d_data_2)

    ifile_1.Close()
    ifile_2.Close()
#------------------
def _save_summary() -> None:
    '''
    Saves Data.d_summary to summary.yaml
    '''

    with open('summary.yaml', 'w', encoding='utf-8') as ofile:
        yaml.dump(Data.d_summary, ofile, indent=2, default_flow_style=False)
#------------------
def main():
    '''
    Script starts here
    '''
    parser = argparse.ArgumentParser(description='Used to validate versions of code that produce potentially different files')
    parser.add_argument('-f', '--files'         , nargs=  2, help='List of files to compare')
    parser.add_argument('-n', '--max_entries'   , type=int , help='Limit running over this number of entries. By default will run over everything', default=-1)
    parser.add_argument('-t', '--max_trees'     , type=int , help='Limit running over this number of trees. By default will run over everything'  , default=-1)
    parser.add_argument('-l', '--log_level'     , type=int , help='Logging level'                              , default=20, choices=[10, 20, 30, 40])
    parser.add_argument('-e', '--exclude'       , nargs='+', help='List of trees that should not be compared'  , default=[], )
    parser.add_argument('-r', '--raise_if_diff' ,            help='If used, will fail as soon as it finds trees with different branches.', action='store_true')

    args = parser.parse_args()

    LogStore.set_level('rx_scripts:compare_files', args.log_level)

    Data.max_entries   = args.max_entries
    Data.max_trees     = args.max_trees
    Data.l_exclude     = args.exclude
    Data.raise_if_diff = args.raise_if_diff

    [file_1, file_2] = args.files

    Data.file_name_1 = file_1
    Data.file_name_2 = file_2

    _validate(file_1, file_2)
    _save_summary()
#------------------
if __name__ == '__main__':
    main()



'''
Module containing plotter class
'''

import numpy
import matplotlib.pyplot as plt

from dmu.logging.log_store import LogStore
from dmu.plotting.plotter  import Plotter

log = LogStore.add_logger('dmu:plotting:Plotter1D')
# --------------------------------------------
class Plotter1D(Plotter):
    '''
    Class used to plot columns in ROOT dataframes
    '''
    # --------------------------------------------
    def __init__(self, d_rdf=None, cfg=None):
        '''
        Parameters:

        d_rdf (dict): Dictionary mapping the kind of sample with the ROOT dataframe
        cfg   (dict): Dictionary with configuration, e.g. binning, ranges, etc
        '''

        super().__init__(d_rdf=d_rdf, cfg=cfg)
    #-------------------------------------
    def _get_labels(self, var : str) -> tuple[str,str]:
        if 'labels' not in self._d_cfg['plots'][var]:
            return var, 'Entries'

        xname, yname = self._d_cfg['plots'][var]['labels' ]

        return xname, yname
    #-------------------------------------
    def _plot_var(self, var):
        '''
        Will plot a variable from a dictionary of dataframes
        Parameters
        --------------------
        var   (str)  : name of column
        '''
        # pylint: disable=too-many-locals

        d_cfg = self._d_cfg['plots'][var]

        minx, maxx, bins = d_cfg['binning']
        yscale           = d_cfg['yscale' ] if 'yscale' in d_cfg else 'linear'
        xname, yname     = self._get_labels(var)

        normalized=False
        if 'normalized' in d_cfg:
            normalized = d_cfg['normalized']

        title = ''
        if 'title'      in d_cfg:
            title = d_cfg['title']

        d_data = {}
        for name, rdf in self._d_rdf.items():
            d_data[name] = rdf.AsNumpy([var])[var]

        if maxx <= minx + 1e-5:
            log.info(f'Bounds not set for {var}, will calculated them')
            minx, maxx = self._find_bounds(d_data = d_data, qnt=minx)
            log.info(f'Using bounds [{minx:.3e}, {maxx:.3e}]')
        else:
            log.debug(f'Using bounds [{minx:.3e}, {maxx:.3e}]')

        l_bc_all = []
        d_wgt    = self._get_weights(var)
        for name, arr_val in d_data.items():
            arr_wgt    = d_wgt[name] if d_wgt is not None else None

            self._print_weights(arr_wgt, var, name)
            l_bc, _, _ = plt.hist(arr_val, weights=arr_wgt, bins=bins, range=(minx, maxx), density=normalized, histtype='step', label=name)
            l_bc_all  += numpy.array(l_bc).tolist()

            plt.yscale(yscale)
            plt.xlabel(xname)
            plt.ylabel(yname)

        if yscale == 'linear':
            plt.ylim(bottom=0)

        max_y = max(l_bc_all)
        plt.ylim(top=1.2 * max_y)
        plt.title(title)
    # --------------------------------------------
    def _plot_lines(self, var : str):
        '''
        Will plot vertical lines for some variables

        var (str) : name of variable
        '''
        if var in ['B_const_mass_M', 'B_M']:
            plt.axvline(x=5280, color='r', label=r'$B^+$'   , linestyle=':')
        elif var == 'Jpsi_M':
            plt.axvline(x=3096, color='r', label=r'$J/\psi$', linestyle=':')
    # --------------------------------------------
    def run(self):
        '''
        Will run plotting
        '''

        fig_size = self._get_fig_size()
        for var in self._d_cfg['plots']:
            log.debug(f'Plotting: {var}')
            plt.figure(var, figsize=fig_size)
            self._plot_var(var)
            self._plot_lines(var)
            self._save_plot(var)
# --------------------------------------------



'''
Script used to print contents of root files
'''

import argparse

from dmu.rfile.rfprinter import RFPrinter

# -----------------------------
class Data:
    '''
    Data class holding shared attributes
    '''
    path   : str
    screen : bool
# -----------------------------
def _get_args():
    parser = argparse.ArgumentParser(description='Script used to print information about ROOT files and dump it to text')
    parser.add_argument('-p', '--path'  , type=str, help='Path to ROOT file')
    parser.add_argument('-s', '--screen', help='If used, will dump output to screen', action='store_true')
    args = parser.parse_args()

    Data.path  = args.path
    Data.screen= args.screen
# -----------------------------
def main():
    '''
    Execution starts here
    '''
    _get_args()
    prt = RFPrinter(path = Data.path)
    prt.save(to_screen = Data.screen)
# -----------------------------
if __name__ == '__main__':
    main()



'''
Module containing utility functions needed by unit tests
'''
import os
from typing              import Union
from dataclasses         import dataclass
from importlib.resources import files

from ROOT import RDF, TFile, RDataFrame

import pandas as pnd
import numpy
import yaml

from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:testing:utilities')
# -------------------------------
@dataclass
class Data:
    '''
    Class storing shared data
    '''
    nentries = 3000
# -------------------------------
def _double_data(d_data : dict) -> dict:
    df_1   = pnd.DataFrame(d_data)
    df_2   = pnd.DataFrame(d_data)

    df     = pnd.concat([df_1, df_2], axis=0)

    d_data = { name : df[name].to_numpy() for name in df.columns }

    return d_data
# -------------------------------
def _add_nans(d_data : dict) -> dict:
    df_good   = pnd.DataFrame(d_data)
    df_bad    = pnd.DataFrame(d_data)
    df_bad[:] = numpy.nan

    df        = pnd.concat([df_good, df_bad])
    d_data    = { name : df[name].to_numpy() for name in df.columns }

    return d_data
# -------------------------------
def get_rdf(kind : Union[str,None] = None,
            repeated : bool        = False,
            add_nans : bool        = False):
    '''
    Return ROOT dataframe with toy data
    '''
    d_data = {}
    if   kind == 'sig':
        d_data['w'] = numpy.random.normal(0, 1, size=Data.nentries)
        d_data['x'] = numpy.random.normal(0, 1, size=Data.nentries)
        d_data['y'] = numpy.random.normal(0, 1, size=Data.nentries)
        d_data['z'] = numpy.random.normal(0, 1, size=Data.nentries)
    elif kind == 'bkg':
        d_data['w'] = numpy.random.normal(1, 1, size=Data.nentries)
        d_data['x'] = numpy.random.normal(1, 1, size=Data.nentries)
        d_data['y'] = numpy.random.normal(1, 1, size=Data.nentries)
        d_data['z'] = numpy.random.normal(1, 1, size=Data.nentries)
    else:
        log.error(f'Invalid kind: {kind}')
        raise ValueError

    if repeated:
        d_data = _double_data(d_data)

    if add_nans:
        d_data = _add_nans(d_data)

    rdf = RDF.FromNumpy(d_data)

    return rdf
# -------------------------------
def get_config(name : Union[str,None] = None):
    '''
    Takes path to the YAML config file, after `dmu_data`
    Returns dictionary with config
    '''
    if name is None:
        raise ValueError('Name not pased')

    cfg_path = files('dmu_data').joinpath(name)
    cfg_path = str(cfg_path)
    with open(cfg_path, encoding='utf-8') as ifile:
        d_cfg = yaml.safe_load(ifile)

    return d_cfg
# -------------------------------
def _get_rdf(nentries : int) -> RDataFrame:
    rdf = RDataFrame(nentries)
    rdf = rdf.Define('x', '0')
    rdf = rdf.Define('y', '1')
    rdf = rdf.Define('z', '2')

    return rdf
# -------------------------------
def get_file_with_trees(path : str) -> TFile:
    '''
    Picks full path to toy ROOT file, in the form of /a/b/c/file.root
    returns handle to it
    '''
    dir_name    = os.path.dirname(path)
    os.makedirs(dir_name, exist_ok=True)

    snap        = RDF.RSnapshotOptions()
    snap.fMode  = 'recreate'

    l_tree_name = ['odir/idir/a', 'dir/b', 'c']
    l_nevt      = [    100, 200, 300]

    l_rdf = [ _get_rdf(nevt) for nevt in l_nevt ]
    for rdf, tree_name in zip(l_rdf, l_tree_name):
        rdf.Snapshot(tree_name, path, ['x', 'y', 'z'], snap)
        snap.fMode  = 'update'

    return TFile(path)



'''
Module containing plotter class
'''

import os
import math
from typing import Union

import numpy
import matplotlib.pyplot as plt

from ROOT                  import RDataFrame
from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:plotting:Plotter')
# --------------------------------------------
class Plotter:
    '''
    Base class of Plotter1D and Plotter2D
    '''
    #-------------------------------------
    def __init__(self, d_rdf=None, cfg=None):
        if not isinstance(  cfg, dict):
            raise ValueError('Config dictionary not passed')

        if not isinstance(d_rdf, dict):
            raise ValueError('Dataframe dictionary not passed')

        self._d_cfg = cfg
        self._d_rdf : dict[str, RDataFrame]    = { name : self._preprocess_rdf(rdf) for name, rdf in d_rdf.items()}
        self._d_wgt : Union[dict[str, Union[numpy.ndarray, None]], None]
    #-------------------------------------
    def _check_quantile(self, qnt : float):
        '''
        Will check validity of quantile
        '''

        if 0.5 < qnt <= 1.0:
            return

        raise ValueError(f'Invalid quantile: {qnt:.3e}, value needs to be in (0.5, 1.0] interval')
    #-------------------------------------
    def _find_bounds(self, d_data : dict, qnt : float = 0.98):
        '''
        Will take dictionary between kinds of data and numpy array
        Will return tuple with bounds, where 95% of the data is found
        '''
        self._check_quantile(qnt)

        l_max = []
        l_min = []

        for arr_val in d_data.values():
            minv = numpy.quantile(arr_val, 1 - qnt)
            maxv = numpy.quantile(arr_val,     qnt)

            l_max.append(maxv)
            l_min.append(minv)

        minx = min(l_min)
        maxx = max(l_max)

        if minx >= maxx:
            raise ValueError(f'Could not calculate bounds correctly: [{minx:.3e}, {maxx:.3e}]')

        return minx, maxx
    #-------------------------------------
    def _preprocess_rdf(self, rdf):
        '''
        rdf (RDataFrame): ROOT dataframe

        returns preprocessed dataframe
        '''

        rdf = self._define_vars(rdf)
        if 'selection' in self._d_cfg:
            rdf = self._apply_selection(rdf)
            rdf = self._max_ran_entries(rdf)

        return rdf
    #-------------------------------------
    def _define_vars(self, rdf):
        '''
        Will define extra columns in dataframe and return updated dataframe
        '''

        if 'definitions' not in self._d_cfg:
            log.debug('No definitions section found, returning same RDF')
            return rdf

        d_def = self._d_cfg['definitions']

        log.info('Defining extra variables')
        for name, expr in d_def.items():
            log.debug(f'{name:<30}{expr:<150}')
            rdf = rdf.Define(name, expr)

        return rdf
    #-------------------------------------
    def _apply_selection(self, rdf):
        '''
        Will take dataframe, apply selection and return dataframe
        '''
        if 'cuts' not in self._d_cfg['selection']:
            log.debug('Cuts not found in selection section, not applying any cuts')
            return rdf

        d_cut = self._d_cfg['selection']['cuts']

        log.info('Applying cuts')
        for name, cut in d_cut.items():
            log.debug(f'{name:<50}{cut:<150}')
            rdf = rdf.Filter(cut, name)

        return rdf
    #-------------------------------------
    def _max_ran_entries(self, rdf):
        '''
        Will take dataframe and randomly drop events
        '''

        if 'max_ran_entries' not in self._d_cfg['selection']:
            log.debug('Cuts not found in selection section, not applying any cuts')
            return rdf

        tot_entries = rdf.Count().GetValue()
        max_entries = self._d_cfg['selection']['max_ran_entries']

        if tot_entries < max_entries:
            log.debug(f'Not dropping dandom entries: {tot_entries} < {max_entries}')
            return rdf

        prescale = math.floor(tot_entries / max_entries)
        if prescale < 2:
            log.debug(f'Not dropping random entries, prescale is below 2: {tot_entries}/{max_entries}')
            return rdf

        rdf = rdf.Filter(f'rdfentry_ % {prescale} == 0', 'max_ran_entries')

        fnl_entries = rdf.Count().GetValue()

        log.info(f'Dropped entries randomly: {tot_entries} -> {fnl_entries}')

        return rdf
    # --------------------------------------------
    def _print_weights(self, arr_wgt : Union[numpy.ndarray, None], var : str, sample : str) -> None:
        if arr_wgt is None:
            log.debug(f'Not using weights for {sample}:{var}')
            return

        num_wgt = len(arr_wgt)
        sum_wgt = numpy.sum(arr_wgt)

        log.debug(f'Using weights [{num_wgt},{sum_wgt:.0f}] for {var}')
    # --------------------------------------------
    def _get_fig_size(self):
        '''
        Will read size list from config dictionary if found
        other wise will return None
        '''
        if 'general' not in self._d_cfg:
            return None

        if 'size' not in self._d_cfg['general']:
            return None

        fig_size = self._d_cfg['general']['size']

        return fig_size
    #-------------------------------------
    def _get_weights(self, var) -> Union[dict[str, Union[numpy.ndarray, None]], None]:
        d_cfg = self._d_cfg['plots'][var]
        if 'weights' not in d_cfg:
            return None

        if hasattr(self, '_d_wgt'):
            return self._d_wgt

        wgt_name = d_cfg['weights']
        d_weight = {sam_name : self._read_weights(wgt_name, rdf) for sam_name, rdf in self._d_rdf.items()}

        self._d_wgt = d_weight

        return d_weight
    # --------------------------------------------
    def _read_weights(self, name : str, rdf : RDataFrame) -> Union[numpy.ndarray, None]:
        v_col = rdf.GetColumnNames()
        l_col = [ col.c_str() for col in v_col ]

        if name not in l_col:
            log.debug(f'Weight {name} not found')
            return None

        arr_wgt = rdf.AsNumpy([name])[name]

        return arr_wgt
    #-------------------------------------
    def _get_plot_name(self, var : str) -> str:
        if 'plots_2d' in self._d_cfg:
            #For 2D plots the name will always be specified in the config
            return var

        if 'name' not in self._d_cfg['plots'][var]:
            # For 1D plots the name can be taken from variable name itself or specified
            return var

        return self._d_cfg['plots'][var]['name']
    #-------------------------------------
    def _save_plot(self, var):
        '''
        Will save to PNG:

        var (str) : Name of variable, needed for plot name
        '''
        plt.legend()

        plt_dir = self._d_cfg['saving']['plt_dir']
        os.makedirs(plt_dir, exist_ok=True)

        name = self._get_plot_name(var)

        plot_path = f'{plt_dir}/{name}.png'
        log.info(f'Saving to: {plot_path}')
        plt.tight_layout()
        plt.savefig(plot_path)
        plt.close(var)
# --------------------------------------------



'''
Module containing utility functions for ML tools
'''

import hashlib
from typing import Union

import numpy
import pandas as pnd

from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:ml:utilities')
# ---------------------------------------------
# Patch dataframe with features
# ---------------------------------------------
def patch_and_tag(df : pnd.DataFrame, value : float = 0) -> pnd.DataFrame:
    '''
    Takes panda dataframe, replaces NaNs with value introduced, by default 0
    Returns array of indices where the replacement happened
    '''
    l_nan = df.index[df.isna().any(axis=1)].tolist()
    nnan  = len(l_nan)
    if nnan == 0:
        log.debug('No NaNs found')
        return df

    log.warning(f'Found {nnan} NaNs, patching them with {value}')

    df_pa = df.fillna(value)

    df_pa.attrs['patched_indices'] = numpy.array(l_nan)

    return df_pa
# ---------------------------------------------
# Cleanup of dataframe with features
# ---------------------------------------------
def cleanup(df : pnd.DataFrame) -> pnd.DataFrame:
    '''
    Takes pandas dataframe with features for classification
    Removes repeated entries and entries with nans
    Returns dataframe
    '''
    df = _remove_repeated(df)
    df = _remove_nans(df)

    return df
# ---------------------------------------------
def _remove_nans(df : pnd.DataFrame) -> pnd.DataFrame:
    if not df.isna().any().any():
        log.debug('No NaNs found in dataframe')
        return df

    ninit = len(df)
    df    = df.dropna()
    nfinl = len(df)

    log.warning(f'NaNs found, cleaning dataset: {ninit} -> {nfinl}')

    return df
# ---------------------------------------------
def _remove_repeated(df : pnd.DataFrame) -> pnd.DataFrame:
    l_hash = get_hashes(df, rvalue='list')
    s_hash = set(l_hash)

    ninit = len(l_hash)
    nfinl = len(s_hash)

    if ninit == nfinl:
        log.debug('No cleaning needed for dataframe')
        return df

    log.warning(f'Repeated entries found, cleaning up: {ninit} -> {nfinl}')

    df['hash_index'] = l_hash
    df               = df.set_index('hash_index', drop=True)
    df_clean         = df[~df.index.duplicated(keep='first')]

    if not isinstance(df_clean, pnd.DataFrame):
        raise ValueError('Cleaning did not return pandas dataframe')

    return df_clean
# ----------------------------------
# ---------------------------------------------
def get_hashes(df_ft : pnd.DataFrame, rvalue : str ='set') -> Union[set, list]:
    '''
    Will return hashes for each row in the feature dataframe

    rvalue (str): Return value, can be a set or a list
    '''

    if   rvalue == 'set':
        res = { hash_from_row(row) for _, row in df_ft.iterrows() }
    elif rvalue == 'list':
        res = [ hash_from_row(row) for _, row in df_ft.iterrows() ]
    else:
        log.error(f'Invalid return value: {rvalue}')
        raise ValueError

    return res
# ----------------------------------
def hash_from_row(row):
    '''
    Will return a hash from a pandas dataframe row
    corresponding to an event
    '''
    l_val   = [ str(val) for val in row ]
    row_str = ','.join(l_val)
    row_str = row_str.encode('utf-8')

    hsh = hashlib.sha256()
    hsh.update(row_str)

    hsh_val = hsh.hexdigest()

    return hsh_val
# ----------------------------------
def index_with_hashes(df):
    '''
    Will:
    - take dataframe with features
    - calculate hashes and add them as the index column
    - drop old index column
    '''

    l_hash = get_hashes(df, rvalue='list')
    ind_hsh= pnd.Index(l_hash)

    df = df.set_index(ind_hsh, drop=True)

    return df
# ----------------------------------



#!/usr/bin/env python3

from dmu.text.transformer import transformer as txt_trf

import argparse
import logging

log = logging.getLogger('dmu_scripts:text:transformer')
#---------------------------------
class data:
    txt = None
    out = None
    cfg = None
    lvl = None
#---------------------------------
def get_args():
    parser=argparse.ArgumentParser(description='Will transform a text file following a set of rules')
    parser.add_argument('-i', '--input' , type=str, help='Path to input file' , required=True) 
    parser.add_argument('-o', '--output', type=str, help='Path to output file, if not passed, it will be same as input, but with trf before extension')
    parser.add_argument('-c', '--config', type=str, help='Path to config file', required=True) 
    parser.add_argument('-l', '--loglvl', type=int, help='Log level'          , default=20, choices=[10, 20, 30, 40]) 
    args = parser.parse_args()

    data.txt = args.input
    data.out = args.output
    data.cfg = args.config
    data.lvl = args.loglvl
#---------------------------------
def set_logs():
    logging.basicConfig()

    log_tr = logging.getLogger('dmu:text:transformer')

    log_tr.setLevel(data.lvl)
    log.setLevel(data.lvl)
#---------------------------------
def main():
    get_args()
    set_logs()

    trf = txt_trf(txt_path=data.txt, cfg_path=data.cfg)
    trf.save_as(data.out)
#---------------------------------
if __name__ == '__main__':
    main()




'''
Module containing RFPrinter
'''
import os

from ROOT import TFile

from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:rfprinter')
#--------------------------------------------------
class RFPrinter:
    '''
    Class meant to print summary of ROOT file
    '''
    #-----------------------------------------
    def __init__(self, path : str):
        '''
        Takes path to root file
        '''
        if not os.path.isfile(path):
            raise FileNotFoundError(f'Cannot find {path}')

        self._root_path = path
        self._text_path = path.replace('.root', '.txt')
    #-----------------------------------------
    def _get_trees(self, ifile):
        '''
        Takes TFile object, returns list of TTree objects
        '''
        l_key=ifile.GetListOfKeys()

        l_tree=[]
        for key in l_key:
            obj=key.ReadObj()
            if obj.InheritsFrom("TTree"):
                fname=ifile.GetName()
                tname=obj.GetName()

                title=f'{fname}/{tname}'
                obj.SetTitle(title)
                l_tree.append(obj)
            elif obj.InheritsFrom("TDirectory"):
                l_tree+=self._get_trees(obj)

        return l_tree
    #---------------------------------
    def _get_tree_info(self, tree):
        '''
        Takes ROOT tree, returns list of strings with information about tree
        '''
        l_branch= tree.GetListOfBranches()
        l_line  = []
        for branch in l_branch:
            bname = branch.GetName()
            leaf  = branch.GetLeaf(bname)
            btype = leaf.GetTypeName()

            l_line.append(f'{"":4}{bname:<100}{btype:<40}')

        return l_line
    #-----------------------------------------
    def _save_info(self, l_info):
        '''
        Takes list of strings, saves it to text file
        '''

        with open(self._text_path, 'w', encoding='utf-8') as ofile:
            for info in l_info:
                ofile.write(f'{info}\n')

        log.info(f'Saved to: {self._text_path}')
    #-----------------------------------------
    def save(self, to_screen=False):
        '''
        Will save a text file with the summary of the ROOT file contents

        to_screen (bool) : If true, will print to screen, default=False
        '''
        l_info = []
        log.info(f'Reading from : {self._root_path}')
        with TFile.Open(self._root_path) as ifile:
            l_tree = self._get_trees(ifile)
            for tree in l_tree:
                l_info+= self._get_tree_info(tree)

        self._save_info(l_info)
        if to_screen:
            for info in l_info:
                log.info(info)
#-----------------------------------------



'''
Module containing generic utility functions
'''
import os
import time
import json
import inspect

from typing import Callable

from functools             import wraps
from dmu.logging.log_store import LogStore

TIMER_ON=False

log = LogStore.add_logger('dmu:generic:utilities')

# --------------------------------
def _get_module_name( fun : Callable) -> str:
    mod = inspect.getmodule(fun)
    if mod is None:
        raise ValueError(f'Cannot determine module name for function: {fun}')

    return mod.__name__
# --------------------------------
def timeit(f):
    '''
    Decorator used to time functions, it is turned off by default, can be turned on with:

    from dmu.generic.utilities import TIMER_ON
    from dmu.generic.utilities import timeit 

    TIMER_ON=True

    @timeit
    def fun():
        ...
    '''
    @wraps(f)
    def wrap(*args, **kw):
        if not TIMER_ON:
            result = f(*args, **kw)
            return result

        ts = time.time()
        result = f(*args, **kw)
        te = time.time()
        mod_nam = _get_module_name(f)
        fun_nam = f.__name__
        log.info(f'{mod_nam}.py:{fun_nam}; Time: {te-ts:.3f}s')

        return result
    return wrap
# --------------------------------
def dump_json(data, path : str, sort_keys : bool = False):
    '''
    Saves data as JSON

    Parameters
    data     : dictionary, list, etc
    path     : Path to JSON file where to save it
    sort_keys: Will set sort_keys argument of json.dump function 
    '''
    dir_name = os.path.dirname(path)
    os.makedirs(dir_name, exist_ok=True)

    with open(path, 'w', encoding='utf-8') as ofile:
        json.dump(data, ofile, indent=4, sort_keys=sort_keys)
# --------------------------------



'''
Module with TrainMva class
'''
import os
from typing import Union

import joblib
import pandas as pnd
import numpy
import matplotlib.pyplot as plt

from sklearn.metrics         import roc_curve, auc
from sklearn.model_selection import StratifiedKFold

from ROOT import RDataFrame

import dmu.ml.utilities    as ut
from dmu.ml.cv_classifier    import CVClassifier as cls
from dmu.plotting.plotter_1d import Plotter1D    as Plotter
from dmu.logging.log_store   import LogStore

log = LogStore.add_logger('data_checks:train_mva')
# ---------------------------------------------
class TrainMva:
    '''
    Interface to scikit learn used to train classifier
    '''
    # ---------------------------------------------
    def __init__(self, bkg=None, sig=None, cfg=None):
        '''
        bkg (ROOT dataframe): Holds real data
        sig (ROOT dataframe): Holds simulation
        cfg (dict)          : Dictionary storing configuration for training
        '''
        if bkg is None:
            raise ValueError('Background dataframe is not a ROOT dataframe')

        if sig is None:
            raise ValueError('Signal dataframe is not a ROOT dataframe')

        if not isinstance(cfg, dict):
            raise ValueError('Config dictionary is not a dictionary')

        self._rdf_bkg = bkg
        self._rdf_sig = sig
        self._cfg     = cfg if cfg is not None else {}

        self._l_model   : cls

        self._l_ft_name = self._cfg['training']['features']

        self._df_ft, self._l_lab = self._get_inputs()
    # ---------------------------------------------
    def _get_inputs(self) -> tuple[pnd.DataFrame, numpy.ndarray]:
        log.info('Getting signal')
        df_sig, arr_lab_sig = self._get_sample_inputs(self._rdf_sig, label = 1)

        log.info('Getting background')
        df_bkg, arr_lab_bkg = self._get_sample_inputs(self._rdf_bkg, label = 0)

        df      = pnd.concat([df_sig, df_bkg], axis=0)
        arr_lab = numpy.concatenate([arr_lab_sig, arr_lab_bkg])

        return df, arr_lab
    # ---------------------------------------------
    def _get_sample_inputs(self, rdf : RDataFrame, label : int) -> tuple[pnd.DataFrame, numpy.ndarray]:
        d_ft = rdf.AsNumpy(self._l_ft_name)
        df   = pnd.DataFrame(d_ft)
        df   = ut.cleanup(df)
        l_lab= len(df) * [label]

        return df, numpy.array(l_lab)
    # ---------------------------------------------
    def _get_model(self, arr_index : numpy.ndarray) -> cls:
        model = cls(cfg = self._cfg)
        df_ft = self._df_ft.iloc[arr_index]
        l_lab = self._l_lab[arr_index]

        log.debug(f'Training feature shape: {df_ft.shape}')
        log.debug(f'Training label size: {len(l_lab)}')

        model.fit(df_ft, l_lab)

        return model
    # ---------------------------------------------
    def _get_models(self):
        # pylint: disable = too-many-locals
        '''
        Will create models, train them and return them
        '''
        nfold = self._cfg['training']['nfold']
        rdmst = self._cfg['training']['rdm_stat']

        kfold = StratifiedKFold(n_splits=nfold, shuffle=True, random_state=rdmst)

        l_model=[]
        ifold=0
        for arr_itr, arr_its in kfold.split(self._df_ft, self._l_lab):
            log.debug(20 * '-')
            log.info(f'Training fold: {ifold}')
            log.debug(20 * '-')
            model = self._get_model(arr_itr)
            l_model.append(model)

            arr_sig_sig_tr, arr_sig_bkg_tr, arr_sig_all_tr, arr_lab_tr = self._get_scores(model, arr_itr, on_training_ok= True)
            arr_sig_sig_ts, arr_sig_bkg_ts, arr_sig_all_ts, arr_lab_ts = self._get_scores(model, arr_its, on_training_ok=False)

            self._plot_scores(arr_sig_sig_tr, arr_sig_sig_ts, arr_sig_bkg_tr, arr_sig_bkg_ts, ifold)

            self._plot_roc(arr_lab_ts, arr_sig_all_ts, arr_lab_tr, arr_sig_all_tr, ifold)

            ifold+=1

        return l_model
    # ---------------------------------------------
    def _get_scores(self, model : cls, arr_index : numpy.ndarray, on_training_ok : bool) -> tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]:
        '''
        Returns a tuple of four arrays

        arr_sig : Signal probabilities for signal
        arr_bkg : Signal probabilities for background
        arr_all : Signal probabilities for both
        arr_lab : Labels for both
        '''
        nentries = len(arr_index)
        log.debug(f'Getting {nentries} signal probabilities')

        df_ft    = self._df_ft.iloc[arr_index]
        arr_prob = model.predict_proba(df_ft, on_training_ok=on_training_ok)
        arr_lab  = self._l_lab[arr_index]

        l_all    = [ sig_prob for [_, sig_prob] in arr_prob ]
        arr_all  = numpy.array(l_all)

        arr_sig, arr_bkg= self._split_scores(arr_prob=arr_prob, arr_label=arr_lab)

        return arr_sig, arr_bkg, arr_all, arr_lab
    # ---------------------------------------------
    def _split_scores(self, arr_prob : numpy.ndarray, arr_label : numpy.ndarray) -> tuple[numpy.ndarray, numpy.ndarray]:
        '''
        Will split the testing scores (predictions) based on the training scores

        tst is a list of lists as [p_bkg, p_sig]
        '''

        l_sig = [ prb[1] for prb, lab in zip(arr_prob, arr_label) if lab == 1]
        l_bkg = [ prb[1] for prb, lab in zip(arr_prob, arr_label) if lab == 0]

        arr_sig = numpy.array(l_sig)
        arr_bkg = numpy.array(l_bkg)

        return arr_sig, arr_bkg
    # ---------------------------------------------
    def _save_model(self, model, ifold):
        '''
        Saves a model, associated to a specific fold
        '''
        model_path = self._cfg['saving']['path']
        if os.path.isfile(model_path):
            log.info(f'Model found in {model_path}, not saving')
            return

        dir_name = os.path.dirname(model_path)
        os.makedirs(dir_name, exist_ok=True)

        model_path = model_path.replace('.pkl', f'_{ifold:03}.pkl')

        log.info(f'Saving model to: {model_path}')
        joblib.dump(model, model_path)
    # ---------------------------------------------
    def _plot_scores(self, arr_sig_trn, arr_sig_tst, arr_bkg_trn, arr_bkg_tst, ifold):
        # pylint: disable = too-many-arguments, too-many-positional-arguments
        '''
        Will plot an array of scores, associated to a given fold
        '''
        log.debug(f'Plotting scores for {ifold} fold')

        if 'val_dir' not in self._cfg['plotting']:
            log.warning('Scores path not passed, not plotting scores')
            return

        val_dir  = self._cfg['plotting']['val_dir']
        val_dir  = f'{val_dir}/fold_{ifold:03}'
        os.makedirs(val_dir, exist_ok=True)

        plt.hist(arr_sig_trn, alpha   =   0.3, bins=50, range=(0,1), color='b', density=True, label='Signal Train')
        plt.hist(arr_sig_tst, histtype='step', bins=50, range=(0,1), color='b', density=True, label='Signal Test')

        plt.hist(arr_bkg_trn, alpha   =   0.3, bins=50, range=(0,1), color='r', density=True, label='Background Train')
        plt.hist(arr_bkg_tst, histtype='step', bins=50, range=(0,1), color='r', density=True, label='Background Test')

        plt.legend()
        plt.title(f'Fold: {ifold}')
        plt.xlabel('Signal probability')
        plt.ylabel('Normalized')
        plt.savefig(f'{val_dir}/scores.png')
        plt.close()
    # ---------------------------------------------
    def _plot_roc(self,
                  l_lab_ts : numpy.ndarray,
                  l_prb_ts : numpy.ndarray,
                  l_lab_tr : numpy.ndarray,
                  l_prb_tr : numpy.ndarray,
                  ifold    : int):
        '''
        Takes the labels and the probabilities and plots ROC
        curve for given fold
        '''
        # pylint: disable = too-many-arguments, too-many-positional-arguments
        log.debug(f'Plotting ROC curve for {ifold} fold')

        val_dir  = self._cfg['plotting']['val_dir']
        val_dir  = f'{val_dir}/fold_{ifold:03}'
        os.makedirs(val_dir, exist_ok=True)

        xval_ts, yval_ts, _ = roc_curve(l_lab_ts, l_prb_ts)
        xval_ts             = 1 - xval_ts
        area_ts             = auc(xval_ts, yval_ts)

        xval_tr, yval_tr, _ = roc_curve(l_lab_tr, l_prb_tr)
        xval_tr             = 1 - xval_tr
        area_tr             = auc(xval_tr, yval_tr)

        min_x = 0
        min_y = 0
        if 'min' in self._cfg['plotting']['roc']:
            [min_x, min_y] = self._cfg['plotting']['roc']['min']

        plt.plot(xval_ts, yval_ts, color='b', label=f'Test: {area_ts:.3f}')
        plt.plot(xval_tr, yval_tr, color='r', label=f'Train: {area_tr:.3f}')
        plt.xlabel('Signal efficiency')
        plt.ylabel('Background efficiency')
        plt.title(f'Fold: {ifold}')
        plt.xlim(min_x, 1)
        plt.ylim(min_y, 1)
        plt.legend()
        plt.savefig(f'{val_dir}/roc.png')
        plt.close()
    # ---------------------------------------------
    def _plot_features(self):
        '''
        Will plot the features, based on the settings in the config
        '''
        d_cfg = self._cfg['plotting']['features']
        ptr   = Plotter(d_rdf = {'Signal' : self._rdf_sig, 'Background' : self._rdf_bkg}, cfg=d_cfg)
        ptr.run()
    # ---------------------------------------------
    def run(self):
        '''
        Will do the training
        '''
        self._plot_features()

        l_mod = self._get_models()
        for ifold, mod in enumerate(l_mod):
            self._save_model(mod, ifold)
# ---------------------------------------------



'''
Module holding zfitter class
'''

import pprint
from typing                   import Union

import numpy
import zfit
import pandas as pd

from scipy                    import stats
from zfit.minimizers.strategy import FailMinimizeNaN
from zfit.result              import FitResult
from zfit.core.data           import Data
from dmu.logging.log_store    import LogStore

log = LogStore.add_logger('dmu:statistics:fitter')
#------------------------------
class FitterGofError(Exception):
    '''
    Exception used when GoF cannot be calculated
    '''
#------------------------------
class FitterFailedFit(Exception):
    '''
    Exception used when fitter fails
    '''
#------------------------------
class Fitter:
    '''
    Class meant to be an interface to underlying fitters
    '''
    # pylint: disable=too-many-instance-attributes
    #------------------------------
    def __init__(self, pdf, data):
        self._data_in = data
        self._pdf     = pdf

        self._data_zf : zfit.data.Data
        self._data_np : numpy.ndarray
        self._obs     : zfit.Space
        self._d_par   : dict

        self._ndof           = 10
        self._pval_threshold = 0.01
        self._initialized    = False
    #------------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._check_data()

        self._intialized = True
    #------------------------------
    def _check_data(self):
        if   isinstance(self._data_in, numpy.ndarray):
            data_np = self._data_in
        elif isinstance(self._data_in, zfit.Data):
            data_np = zfit.run(zfit.z.unstack_x(self._data_in)) # convert original data to numpy array, needed by _calc_gof
        elif isinstance(self._data_in, pd.DataFrame):
            data_np = self._data_in.to_numpy()
        elif isinstance(self._data_in, pd.Series):
            self._data_in = pd.DataFrame(self._data_in)
            data_np = self._data_in.to_numpy()
        else:
            data_type = str(type(self._data_in))
            raise ValueError(f'Data is not a numpy array, zfit.Data or pandas.DataFrame, but {data_type}')

        data_np       = self._check_numpy_data(data_np)
        self._data_np = data_np
        if not isinstance(self._data_in, zfit.Data):
            self._data_zf = zfit.Data.from_numpy(obs=self._pdf.space, array=data_np)
        else:
            self._data_zf = self._data_in
    #------------------------------
    def _check_numpy_data(self, data):
        shp = data.shape
        if   len(shp) == 1:
            pass
        elif len(shp) == 2:
            _, jval = shp
            if jval != 1:
                log.error(f'Invalid data shape: {shp}')
                raise
        else:
            log.error(f'Invalid data shape: {shp}')
            raise

        ival = data.shape[0]

        data = data[~numpy.isnan(data)]
        data = data[~numpy.isinf(data)]

        fval = data.shape[0]

        if ival != fval:
            log.warning(f'Data was trimmed for inf and nan: {ival} -> {fval}')

        return data
    #------------------------------
    def _bin_pdf(self, nbins):
        [[min_x]], [[max_x]] = self._pdf.space.limits
        _, arr_edg = numpy.histogram(self._data_np, bins = nbins, range=(min_x, max_x))

        size = arr_edg.size

        l_bc = []
        for i_edg in range(size - 1):
            low = arr_edg[i_edg + 0]
            hig = arr_edg[i_edg + 1]

            var = self._pdf.integrate(limits = [low, hig])
            val = var.numpy()[0]
            l_bc.append(val * self._data_np.size)

        return numpy.array(l_bc)
    #------------------------------
    def _get_binning(self):
        min_x = numpy.min(self._data_np)
        max_x = numpy.max(self._data_np)
        nbins = self._ndof + self._get_float_pars()

        return nbins, min_x, max_x
    #------------------------------
    def _calc_gof(self):
        log.debug('Calculating GOF')
        nbins, min_x, max_x = self._get_binning()

        log.debug(f'Nbins: {nbins}')
        log.debug(f'Range: [{min_x:.3f}, {max_x:.3f}]')

        arr_data, _ = numpy.histogram(self._data_np, bins = nbins, range=(min_x, max_x))
        arr_data    = arr_data.astype(float)
        arr_modl    = self._bin_pdf(nbins)
        norm        = numpy.sum(arr_data) / numpy.sum(arr_modl)
        arr_modl    = norm * arr_modl
        arr_res     = arr_modl - arr_data

        arr_chi2    = numpy.divide(arr_res ** 2, arr_data, out=numpy.zeros_like(arr_data), where=arr_data!=0)
        sum_chi2    = numpy.sum(arr_chi2)
        pvalue      = 1 - stats.chi2.cdf(sum_chi2, self._ndof)

        log.debug(f'{"Data":<20}{"Model":<20}{"chi2":<20}')
        if pvalue < self._pval_threshold:
            for data, modl, chi2 in zip(arr_data, arr_modl, arr_chi2):
                log.debug(f'{data:<20.0f}{modl:<20.3f}{chi2:<20.3f}')

        log.debug(f'Chi2: {sum_chi2:.3f}')
        log.debug(f'Ndof: {self._ndof}')
        log.debug(f'pval: {pvalue:<.3e}')

        return (sum_chi2, self._ndof, pvalue)
    #------------------------------
    def _get_float_pars(self):
        npar     = 0
        s_par    = self._pdf.get_params()
        for par in s_par:
            if par.floating:
                npar+=1

        self._d_par = {par.name : par for par in s_par}

        return npar
    #------------------------------
    def _reshuffle_pdf_pars(self):
        '''
        Will move floating parameters of PDF according
        to uniform PDF
        '''

        s_par = self._pdf.get_params(floating=True)
        for par in s_par:
            ival = par.value()
            fval = numpy.random.uniform(par.lower, par.upper)
            par.set_value(fval)
            log.debug(f'{par.name:<20}{ival:<15.3f}{"->":<10}{fval:<15.3f}{"in":<5}{par.lower:<15.3e}{par.upper:<15.3e}')
    #------------------------------
    def _set_pdf_pars(self, res):
        '''
        Will set the PDF floating parameter values as the result instance
        '''
        l_par_flt = list(self._pdf.get_params(floating= True))
        l_par_fix = list(self._pdf.get_params(floating=False))
        l_par     = l_par_flt + l_par_fix

        d_val = { par.name : dc['value'] for par, dc in res.params.items()}

        log.debug('Setting PDF parameters to best result')
        for par in l_par:
            if par.name not in d_val:
                log.debug(f'Skipping {par.name} = {par.value().numpy():.3e}')
                continue

            val = d_val[par.name]
            log.debug(f'{"":<4}{par.name:<20}{"->":<10}{val:<20.3e}')
            par.set_value(val)
    #------------------------------
    def _get_constraints(self, cfg : dict) -> list[zfit.constraint.GaussianConstraint]:
        '''
        Takes dictionary of constraints as floats, returns list of GaussianConstraint objects
        '''
        if 'constraints' not in cfg:
            log.debug('Not using any constraint')
            return []

        d_const = cfg['constraints']
        s_par   = self._pdf.get_params(floating=True)
        d_par   = { par.name : par for par in s_par}

        log.info('Adding constraints:')
        l_const = []
        for par_name, (par_mu, par_sg) in d_const.items():
            if par_name not in d_par:
                log.error(s_par)
                raise ValueError(f'Parameter {par_name} not found among floating parameters of model, above')

            par = d_par[par_name]

            if par_sg == 0:
                par.floating = False
                log.info(f'{"":<4}{par_name:<15}{par_mu:<15.3e}{par_sg:<15.3e}')
                continue

            const = zfit.constraint.GaussianConstraint(params=par, observation=float(par_mu), uncertainty=float(par_sg))
            log.info(f'{"":<4}{par_name:<25}{par_mu:<15.3e}{par_sg:<15.3e}')
            l_const.append(const)

        return l_const
    #------------------------------
    def _get_ranges(self, cfg : dict) -> list:
        if 'ranges' not in cfg:
            return [None]

        ranges = cfg['ranges']
        log.info('-' * 30)
        log.info(f'{"Low edge":>15}{"High edge":>15}')
        log.info('-' * 30)
        for rng in ranges:
            log.info(f'{rng[0]:>15.3e}{rng[1]:>15.3e}')

        return ranges
    #------------------------------
    def _get_subdataset(self, cfg : dict) -> Data:
        if 'nentries' not in cfg:
            return self._data_zf

        nentries_out = cfg['nentries']
        arr_inp      = self._data_zf.to_numpy().flatten()
        nentries_inp = len(arr_inp)
        if nentries_inp <= nentries_out:
            log.warning(f'Input dataset in smaller than output dataset, {nentries_inp} < {nentries_out}')
            return self._data_zf

        has_weights = self._data_zf.weights is not None

        if has_weights:
            arr_wgt = self._data_zf.weights.numpy()
            arr_inp = numpy.array([arr_inp, arr_wgt]).T

        arr_out = numpy.random.choice(arr_inp, size=nentries_out, replace=False)
        if has_weights:
            arr_out = arr_out.T[0]
            arr_wgt = arr_out.T[1]
        else:
            arr_wgt = None

        data = zfit.data.from_numpy(array=arr_out, weights=arr_wgt, obs=self._data_zf.obs)

        return data
    #------------------------------
    def _get_binned_observable(self, nbins : int):
        obs = self._pdf.space
        [[minx]], [[maxx]] = obs.limits

        binning = zfit.binned.RegularBinning(nbins, minx, maxx, name=obs.label)
        obs_bin = zfit.Space(obs.label, binning=binning)

        return obs_bin
    #------------------------------
    def _get_nbins(self, cfg : dict) -> Union[None, int]:
        if 'likelihood' not in cfg:
            return None

        if 'nbins' not in cfg['likelihood']:
            return None

        return cfg['likelihood']['nbins']
    #------------------------------
    def _get_nll(self, data_zf, constraints, frange, cfg):
        nbins     = self._get_nbins(cfg)
        if nbins is None:
            log.info('No binning was specified, will do unbinned fit')
            pdf = self._pdf
        else:
            log.info(f'Using {nbins} bins for fit')
            obs     = self._get_binned_observable(nbins)
            pdf     = zfit.pdf.BinnedFromUnbinnedPDF(self._pdf, obs)
            data_zf = data_zf.to_binned(obs)

        if not self._pdf.is_extended and nbins is None:
            nll = zfit.loss.UnbinnedNLL(        model=pdf, data=data_zf, constraints=constraints, fit_range=frange)
            return nll

        if     self._pdf.is_extended and nbins is None:
            nll = zfit.loss.ExtendedUnbinnedNLL(model=pdf, data=data_zf, constraints=constraints, fit_range=frange)
            return nll

        if frange is not None:
            raise ValueError('Fit range cannot be defined for binned likelihoods')

        if not self._pdf.is_extended:
            nll = zfit.loss.BinnedNLL(          model=pdf, data=data_zf, constraints=constraints)
            return nll

        if     self._pdf.is_extended:
            nll = zfit.loss.ExtendedBinnedNLL(  model=pdf, data=data_zf, constraints=constraints)
            return nll

        raise ValueError('Likelihood was neither Binned nor Unbinned nor Extended nor non-extended')
    #------------------------------
    def _get_full_nll(self, cfg : dict):
        constraints = self._get_constraints(cfg)
        ranges      = self._get_ranges(cfg)
        data_zf     = self._get_subdataset(cfg)
        l_nll       = [ self._get_nll(data_zf, constraints, frange, cfg) for frange in ranges ]
        nll         = sum(l_nll[1:], l_nll[0])

        return nll
    #------------------------------
    def _print_pars(self, cfg : dict):
        '''
        Will print current values parameters in cfg['print_pars'] list, if present
        '''

        if 'print_pars' not in cfg:
            return

        l_par_name = cfg['print_pars']
        d_par_val  = { name : par.value().numpy() for name, par in self._d_par.items() if name in l_par_name}

        l_name = list(d_par_val.keys())
        l_value= list(d_par_val.values())

        l_form = [   f'{var:<10}' for var in l_name]
        header = ''.join(l_form)

        l_form = [f'{val:<10.3f}' for val in l_value]
        parval = ''.join(l_form)

        log.info(header)
        log.info(parval)
    #------------------------------
    def _minimize(self, nll, cfg : dict) -> tuple[FitResult, tuple]:
        mnm = zfit.minimize.Minuit()
        res = mnm.minimize(nll)

        try:
            gof = self._calc_gof()
        except FitterGofError as exc:
            raise FitterGofError('Cannot calculate GOF') from exc

        chi2, _, pval  = gof
        stat           = res.status

        log.info(f'{chi2:<10.3f}{pval:<10.3e}{stat:<10}')
        self._print_pars(cfg)

        return res, gof
    #------------------------------
    def _fit_retries(self, cfg : dict) -> tuple[dict, FitResult]:
        ntries       = cfg['strategy']['retry']['ntries']
        pvalue_thresh= cfg['strategy']['retry']['pvalue_thresh']
        ignore_status= cfg['strategy']['retry']['ignore_status']

        nll        = self._get_full_nll(cfg = cfg)
        d_pval_res = {}
        last_res   = None
        for i_try in range(ntries):
            try:
                res, gof = self._minimize(nll, cfg)
            except (FailMinimizeNaN, FitterGofError, RuntimeError):
                self._reshuffle_pdf_pars()
                log.warning(f'Fit {i_try:03}/{ntries:03} failed due to exception')
                continue

            last_res = res
            bad_fit  = res.status != 0 or not res.valid

            if not ignore_status and bad_fit:
                self._reshuffle_pdf_pars()
                log.info(f'Fit {i_try:03}/{ntries:03} failed, status/validity: {res.status}/{res.valid}')
                continue

            chi2, _, pval   = gof
            d_pval_res[chi2]=res

            if pval > pvalue_thresh:
                log.info(f'Reached {pval:.3f} (> {pvalue_thresh:.3f}) threshold after {i_try + 1} attempts')
                return {chi2 : res}, res

            self._reshuffle_pdf_pars()

        if last_res is None:
            raise FitterFailedFit('Cannot find any valid fit')

        return d_pval_res, last_res
    #------------------------------
    def _pick_best_fit(self, d_pval_res : dict, last_res : FitResult) -> FitResult:
        nsucc = len(d_pval_res)
        if nsucc == 0:
            log.warning('None of the fits succeeded, returning last result')
            self._set_pdf_pars(last_res)

            return last_res

        l_pval_res= list(d_pval_res.items())
        l_pval_res.sort()
        _, res = l_pval_res[0]

        log.debug('Picking out best fit from {nsucc} fits')
        for chi2, _ in l_pval_res:
            log.debug(f'{chi2:.3f}')

        self._set_pdf_pars(res)

        return res
    #------------------------------
    def _fit_in_steps(self, cfg : dict) -> FitResult:
        l_nsample = cfg['strategy']['steps']['nsteps']
        l_nsigma  = cfg['strategy']['steps']['nsigma']
        l_yield   = cfg['strategy']['steps']['yields']

        res = None
        for nsample, nsigma in zip(l_nsample, l_nsigma):
            log.info(f'Fitting with {nsample} samples')
            cfg_step             = dict(cfg)
            cfg_step['nentries'] = nsample

            nll    = self._get_full_nll(cfg = cfg_step)
            res, _ = self._minimize(nll, cfg_step)
            res.hesse(method='minuit_hesse')
            self._update_par_bounds(res, nsigma=nsigma, yields=l_yield)

        log.info('Fitting full sample')
        nll    = self._get_full_nll(cfg = cfg)
        res, _ = self._minimize(nll, cfg)
        res.hesse(method='minuit_hesse')

        if res is None:
            nsteps = len(l_nsample)
            raise ValueError(f'No fit out of {nsteps} was done')

        return res
    #------------------------------
    def _result_to_value_error(self, res : FitResult) -> dict[str, list[float]]:
        d_par = {}
        for par, d_val in res.params.items():
            try:
                val = d_val['value']
                err = d_val['hesse']['error']
            except KeyError as exc:
                pprint.pprint(d_val)
                raise KeyError(f'Cannot extract value, hesse or error from dictionary above') from exc

            d_par[par.name] = [val, err]

        return d_par
    #------------------------------
    def _update_par_bounds(self, res : FitResult, nsigma : float, yields : list[str]) -> None:
        s_shape_par = self._pdf.get_params(is_yield=False, floating=True)
        d_shp_par   = { par.name : par for par in s_shape_par if par.name not in yields}
        d_fit_par   = self._result_to_value_error(res)

        log.info(60 * '-')
        log.info(f'{"Parameter":<20}{"Low bound":<20}{"High bound":<20}')
        log.info(60 * '-')
        for name, [val, err] in d_fit_par.items():
            if name not in d_shp_par:
                log.debug(f'Skipping {name} parameter')
                continue

            shape       = d_shp_par[name]
            shape.lower = val - nsigma * err
            shape.upper = val + nsigma * err

            log.info(f'{name:<20}{val - err:<20.3e}{val + err:<20.3e}')
    #------------------------------
    def fit(self, cfg : Union[dict, None] = None):
        '''
        Runs the fit using the configuration specified by the cfg dictionary

        Returns fit result
        '''
        self._initialize()

        cfg = {} if cfg is None else cfg

        log.info(f'{"chi2":<10}{"pval":<10}{"stat":<10}')
        if 'strategy' not in cfg:
            nll    = self._get_full_nll(cfg = cfg)
            res, _ = self._minimize(nll, cfg)
            res.hesse(method='minuit_hesse')
        elif 'retry' in cfg['strategy']:
            d_pval_res, last_res = self._fit_retries(cfg)
            res = self._pick_best_fit(d_pval_res, last_res)
        elif 'steps' in cfg['strategy']:
            res = self._fit_in_steps(cfg)
        else:
            raise ValueError('Unsupported fitting strategy')


        return res
#------------------------------



'''
Module used to hold transformer class
'''

import os
import pprint

import toml
import numpy

from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:text:transformer')
# -------------------------------------------------------------------------------------------
class transformer:
    # pyling disable = invalid-name
    '''
    Class used to apply transformations to text files
    '''
    # -----------------------------------------
    def __init__(self, txt_path=None, cfg_path=None):
        '''
        txt_path (str): Path to text file to be transformed, can have any extension, py, txt, log, etc
        cfg_path (str): Path to TOML file holding configuration needed for transformations
        '''
        self._txt_path = txt_path
        self._cfg_path = cfg_path
        self._suffix   = 'trf'

        self._l_line   = None
        self._cfg      = None

        self._initialized = False
    # -----------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._check_file(self._txt_path)
        self._check_file(self._cfg_path)
        self._load_input()
        self._cfg = toml.load(self._cfg_path)

        self._initialized=True
    # -----------------------------------------
    def _check_file(self, file_path):
        '''
        Will raise exception if path not found

        file_path (str): path to file
        '''
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f'File not found: {file_path}')

        log.debug(f'Found: {file_path}')
    # -----------------------------------------
    def _load_input(self):
        '''
        Will open  self._txt_path and put the lines in self._l_line
        '''
        with open(self._txt_path) as ifile:
            self._l_line = ifile.read().splitlines()

            nline = len(self._l_line)
            log.info(f'Found {nline} lines in {self._txt_path}')
    # -----------------------------------------
    def _get_out_path(self, out_path):
        '''
        Will return name of output file
        If arg is not None, will make directory (in case it does not exist) and return arg
        If arg is None, will rename input path using suffix  and return
        '''
        if out_path is not None:
            dir_name = os.path.dirname(out_path)
            os.makedirs(dir_name, exist_ok=True)

            return out_path

        file_name = os.path.basename(self._txt_path)
        if '.' not in file_name:
            return f'{file_name}_{self._suffix}'

        l_part     = file_name.split('.')
        bef_ext    = l_part[-2]
        l_part[-2] = f'{bef_ext}_{self._suffix}'

        file_name  = '.'.join(l_part)
        file_dir   = os.path.dirname(self._txt_path)

        return f'{file_dir}/{file_name}'
    # -----------------------------------------
    def _transform(self, l_line, trf):
        log.info(f'{"":<4}{trf}')

        if trf == 'append':
            return self._apply_append(l_line)
        else:
            raise ValueError(f'Invalid transformation: {trf}')

        return l_line
    # -----------------------------------------
    def _apply_append(self, l_line):
        '''
        Will take list of lines
        and return list of lines with extra lines appended
        according to config file
        '''
        d_append = self._cfg['trf']['append']

        for target, l_to_be_added in d_append.items():
            l_to_be_added = self._format_lines(l_to_be_added)
            arr_line      = numpy.array(self._l_line)
            arr_index,    = numpy.where(self._find_append_index(arr_line, target))

            if arr_index.size  == 0:
                pprint.pprint(self._l_line)
                raise RuntimeError(f'No instance of \"{target}\" found in \"{self._txt_path}\"')

            for index in arr_index:
                org_line      = l_line[index]
                ext_line      = '\n'.join(l_to_be_added)
                l_line[index] = f'{org_line}\n{ext_line}'

        return l_line
    # -----------------------------------------
    def _find_append_index(self, l_line, target):
        '''
        Returns list of flags denoting if target was or not fouund in list l_line
        target can be exact or included in the l_line elements
        '''
        is_subst = False
        try:
            is_subst = self._cfg['settings']['as_substring']
        except:
            pass

        if not is_subst:
            log.debug(f'Searching exact matches for target: {target}')
            l_flag = [ target == element for element in l_line ]
        else:
            log.debug(f'Searching with substrings for target: {target}')
            l_flag = [ target in element for element in l_line ]

        return l_flag
    # -----------------------------------------
    def _format_lines(self, l_line):
        '''
        If format was specified in the settings section, will format the
        elements of the input list of lines
        '''
        if 'settings' not in self._cfg:
            return l_line

        if 'format'   not in self._cfg['settings']:
            return l_line

        fmt         = self._cfg['settings']['format']
        l_formatted = [ fmt.format(line) for line in l_line ]

        return l_formatted
    # -----------------------------------------
    def save_as(self, out_path=None):
        '''
        Saves text file after transformation to `out_path`
        If no path is passed, will name as:

        /some/dir/file.txt -> /some/dir/file_trf.txt
        '''
        self._initialize()

        log.info(20 * '-')
        log.info('Applying transformations')
        log.info(20 * '-')
        for trf in  self._cfg['trf']:
            self._l_line = self._transform(self._l_line, trf)

        out_path = self._get_out_path(out_path)
        log.info(f'Saving to: {out_path}')
        with open(out_path, 'w') as ofile:
            text = '\n'.join(self._l_line)
            ofile.write(text)
# -------------------------------------------------------------------------------------------



'''
Module containing the Function class
'''
import os
import json

from typing import Any

import numpy
import matplotlib.pyplot as plt

from scipy.interpolate     import interp1d
from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:stats:function')
#---------------------------------------------------------
class FunOutOfBounds(Exception):
    '''
    Will be raised when function defined between [a, b] is evaluated outside
    '''
#---------------------------------------------------------
class Function:
    '''
    Class meant to represent a 1D function created from (x, y) coordinates
    '''
    #------------------------------------------------
    def __init__(self, x : list | numpy.ndarray, y : list | numpy.ndarray, kind : str = 'cubic'):
        '''
        x (list) : List with x coordinates
        y (list) : List with y coordinates
        '''

        x = self._array_to_list(x)
        y = self._array_to_list(y)

        if len(x) != len(y):
            raise ValueError('X and Y coordinates have different lengths')

        npoint = len(x)
        if npoint < 4:
            raise ValueError('Need at least four points, found {npoint}')

        x, y = self._remove_duplicates(x=x, y=y)

        self._max_entries = 400
        self._l_x = x
        self._l_y = y
        self._kind= kind
        self._tag = 'no_tag'

        self._interpolator = interp1d(self._l_x, self._l_y, kind=self._kind)

        self._update_data()
    #------------------------------------------------
    def __eq__(self, othr):
        if not isinstance(othr, Function):
            log.warning('Comparison not done with instance of Function')
            return False

        d_self = self.__dict__
        d_othr = othr.__dict__

        if '_interpolator' in d_self:
            del d_self['_interpolator']

        if '_interpolator' in d_othr:
            del d_othr['_interpolator']

        return d_self == d_othr
    #------------------------------------------------
    def __str__(self):
        npoints = len(self._l_x)
        max_x   = max(self._l_x)
        min_x   = min(self._l_x)

        max_y   = max(self._l_y)
        min_y   = min(self._l_y)

        line = f'\n{"Points":<20}{npoints:<20}\n'
        line+= '-------------------------\n'
        line+= f'{"x-max":<20}{max_x:<20}\n'
        line+= f'{"x-min":<20}{min_x:<20}\n'
        line+= f'{"y-max":<20}{max_y:<20}\n'
        line+= f'{"y-min":<20}{min_y:<20}'

        return line
    #------------------------------------------------
    def __call__(self, xval : float | numpy.ndarray | list, off_bounds_raise : bool = False) -> numpy.ndarray:
        '''
        Class taking value of x coordinates as a float, numpy array or list
        It will interpolate y value and return value
        '''
        if not off_bounds_raise:
            xval = self._push_in_bounds(xval)

        self._check_xval_validity(xval)

        return self._interpolator(xval)
    #------------------------------------------------
    def _push_in_bounds(self, xval : float | numpy.ndarray | list) -> numpy.ndarray:
        '''
        If the xval container, has elements above (below) the upper (lower) bound, these events will be set to the closest bound
        '''

        xval = numpy.array(xval).flatten().astype(float)

        max_x = max(self._l_x)
        min_x = min(self._l_x)

        if ((min_x <= xval) & (xval <= max_x)).all():
            log.debug('Input array within bounds, will not push elements')
            return xval


        xmod = numpy.clip(xval, min_x, max_x)

        arr_diff = xval != xmod
        arr_indx = numpy.where(arr_diff)[0]
        ndiff    = numpy.sum(arr_diff)
        arr_indx = arr_indx[:20]

        log.warning(f'Sending {ndiff} entries inside bounds [{min_x:.3e}, {max_x:.3e}]')

        for indx in arr_indx:
            org = xval[indx]
            mod = xmod[indx]

            log.info(f'{org:<20.5e}{"-->":<20}{mod:<20.5}')

        return xmod
    #------------------------------------------------
    @staticmethod
    def json_decoder(d_attr):
        '''
        Takes dictionary of attributes from JSON serialization
        Returns instance of Function
        '''

        if '_l_x' not in d_attr:
            raise KeyError('X values not found')

        if '_l_y' not in d_attr:
            raise KeyError('Y values not found')

        if '_tag' not in d_attr:
            raise KeyError('tag not found')

        x    = d_attr['_l_x' ]
        y    = d_attr['_l_y' ]
        kind = d_attr['_kind']
        tag  = d_attr['_tag' ]

        fun  = Function(x=x, y=y, kind=kind)
        fun.tag = tag

        return fun
    #------------------------------------------------
    @property
    def tag(self):
        '''
        Returns string simbolyzing tag of function
        '''
        return self._tag

    @tag.setter
    def tag(self, value : str):
        '''
        This sets the _tag property of the function
        '''
        self._tag = value
    #------------------------------------------------
    @staticmethod
    def load(path : str):
        '''
        Will take path to JSON file with serialized function
        Will return function instance
        '''

        if not os.path.isfile(path):
            raise FileNotFoundError(f'Cannot find: {path}')

        with open(path, encoding='utf-8') as ifile:
            fun = json.loads(ifile.read(), object_hook=Function.json_decoder)

        log.info(f'Loaded from: {path}')

        return fun
    #------------------------------------------------
    def _array_to_list(self, x : Any):
        '''
        Transform from ndarray to list
        Return x if already list
        Raise otherwise
        '''
        if isinstance(x, list):
            log.debug('Already found list')
            return x

        if isinstance(x, numpy.ndarray):
            log.debug('Transforming argument to list')
            return x.tolist()

        raise ValueError('Object introduced is neither a list nor a numpy array')
    #------------------------------------------------
    def _update_data(self):
        '''
        If number of entries in dataset is larger than _max_entries:

        Use interpolator to scan function and get new (x, y) pairs.
        '''
        norg = len(self._l_x)
        if norg <= self._max_entries:
            return

        log.info(f'Trimming dataset: {norg} -> {self._max_entries}')

        min_x = min(self._l_x)
        max_x = max(self._l_x)

        arr_x = numpy.linspace(min_x, max_x, self._max_entries)
        arr_y = self(arr_x)

        self._l_x = arr_x.tolist()
        self._l_y = arr_y.tolist()
    #------------------------------------------------
    def _remove_duplicates(self, x : list, y : list):
        '''
        Takes two lists with the same sizes and remove (x, y) points with repeated
        x coordinates.
        Return tuple with x and y after removal
        '''

        norg  = len(x)

        d_tmp = dict(zip(x, y))

        x = list(d_tmp.keys())
        y = list(d_tmp.values())

        nfnl  = len(x)

        if norg != nfnl:
            log.warning(f'Found duplicates: {norg} -> {nfnl}')

        return x, y
    #------------------------------------------------
    def _check_xval_validity(self, xval : float | numpy.ndarray | list):
        '''
        Will check that xval is an acceptable value for the function to be evaluated at
        '''

        if isinstance(xval, list):
            xval = numpy.array(xval)

        if not isinstance(xval, (float, numpy.ndarray)):
            raise ValueError(f'x value is not a float or numpy array: {xval}')

        check_within_bounds_vect = numpy.vectorize(self._check_within_bounds)
        check_within_bounds_vect(xval)
    #------------------------------------------------
    def _check_within_bounds(self, xval : float):
        '''
        Check that xval is within bounds of function
        '''

        if xval < min(self._l_x) or xval > max(self._l_x):
            print(self)
            raise FunOutOfBounds(f'x value outside bounds: {xval}')
    #------------------------------------------------
    def _json_encoder(self, obj):
        '''
        Takes Function object
        Returns dictionary of attributes for encoding
        '''
        d_data = obj.__dict__

        if '_interpolator' in d_data:
            del d_data['_interpolator']

        return d_data
    #------------------------------------------------
    def _save_plot(self, path : str):
        '''
        Takes path to PNG, saves scatter plot of l_y vs l_x
        '''

        plt.plot(self._l_x, self._l_y)
        plt.savefig(path)
        plt.close()

        log.info(f'Saved to: {path}')
    #------------------------------------------------
    def save(self, path : str, plot : bool = False):
        '''
        Saves current object to JSON

        path (str): Path to file, needs to end in .json
        '''

        if not path.endswith('.json'):
            raise ValueError(f'Output path does not end in .json: {path}')

        dir_name = os.path.dirname(path)
        os.makedirs(dir_name, exist_ok=True)

        with open(path, 'w', encoding='utf-8') as ofile:
            json.dump(self, ofile, indent=4, default=self._json_encoder)

        if plot:
            path = path.replace('.json', '.png')
            self._save_plot(path)

        log.info(f'Saved to: {path}')
#------------------------------------------------



'''
Script used to implement connection to servers
'''

import os
import copy
import argparse

import yaml
from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:scripts:coned')
#---------------------------------------
class Data:
    '''
    Class used to store shared data
    '''
    logl : int
    dry  : bool
    prnt : bool
    cfg  : dict
    l_ad : list[str]
    l_rm : list[str]
#----------------------------
def _print_configs():
    '''
    Prints configuration
    '''

    yaml_output = yaml.dump(Data.cfg, default_flow_style=False)
    print(yaml_output)
#----------------------------
def _initialize():
    _load_config()

    LogStore.set_level('dmu:scripts:coned', Data.logl)

    log.debug(f'Running at {Data.logl} logging level')
#----------------------------
def _get_args():
    '''
    Will parse arguments
    '''
    parser = argparse.ArgumentParser(description='Used to edit and print server list specified by ~/.config/connect/servers.yaml')
    parser.add_argument('-p', '--print'  , help ='Prints config settings and exits', action='store_true')
    parser.add_argument('-l', '--log_lvl', type =int, help='Logging level', default=20, choices=[10,20,30])
    parser.add_argument('-a', '--add'    , nargs=3  , help='Adds task to given server, e.g. task 123 server'     , default=[])
    parser.add_argument('-r', '--rem'    , nargs=3  , help='Removes task from given server, e.g. task 123 server', default=[])
    parser.add_argument('-d', '--dry'    , help='Run dry run, for adding and removing entries', action='store_true')
    args   = parser.parse_args()

    Data.prnt = args.print
    Data.logl = args.log_lvl
    Data.l_ad = args.add
    Data.l_rm = args.rem
    Data.dry  = args.dry
#---------------------------------------
def _load_config():
    home_dir    = os.environ['HOME']
    config_path = f'{home_dir}/.config/dmu/ssh/servers.yaml'
    if not os.path.isfile(config_path):
        raise FileNotFoundError(f'Config not found: {config_path}')

    with open(config_path, encoding='utf-8') as ifile:
        Data.cfg = yaml.safe_load(ifile)
#---------------------------------------
def _dump_config(cfg : dict):
    if cfg == Data.cfg:
        log.debug('Config was not modified, will not save it')
        return

    home_dir    = os.environ['HOME']
    config_path = f'{home_dir}/.config/dmu/ssh/servers.yaml'
    if not os.path.isfile(config_path):
        raise FileNotFoundError(f'Config not found: {config_path}')

    if Data.dry:
        content = yaml.dump(cfg, default_flow_style=False)
        print(content)
        return

    with open(config_path, 'w', encoding='utf-8') as ofile:
        yaml.dump(cfg, ofile, default_flow_style=False)
#---------------------------------------
def _get_updated_config() -> dict:
    log.debug('Getting updated config')

    cfg = copy.deepcopy(Data.cfg)
    cfg = _add_task(cfg)
    cfg = _remove_task(cfg)

    return cfg
#---------------------------------------
def _add_task(cfg : dict) -> dict:
    if len(Data.l_ad) == 0:
        log.debug('No task added')
        return cfg

    [task, machine, server] = Data.l_ad
    if server not in cfg:
        cfg[server] = {}

    if machine not in cfg[server]:
        cfg[server][machine] = []

    cfg[server][machine].append(task)

    log.info(f'{"Added":<10}{server:<20}{machine:<10}{task:<20}')

    return cfg
#---------------------------------------
def _remove_task(cfg : dict) -> dict:
    if len(Data.l_rm) == 0:
        log.debug('No task removed')
        return cfg

    [task, machine, server] = Data.l_rm
    if server not in cfg:
        log.warning(f'Server {server} not found')
        return cfg

    if machine not in cfg[server]:
        log.warning(f'Machine {machine} not found in server {server}')
        return cfg

    l_task = cfg[server][machine]
    if task not in l_task:
        log.warning(f'Task {task} not found in {server}:{machine}')
        return cfg

    index = l_task.index(task)
    del l_task[index]
    cfg[server][machine] = l_task

    log.info(f'{"Removed":<10}{server:<20}{machine:<10}{task:<20}')

    cfg = _trim_config(cfg, machine, server)

    return cfg
#---------------------------------------
def _trim_config(cfg : dict, machine : str, server : str) -> dict:
    if cfg[server][machine] == []:
        log.debug(f'Trimming {server}:{machine}')
        del cfg[server][machine]

    if cfg[server] == {}:
        log.debug(f'Trimming {server}')
        del cfg[server]

    return cfg
#---------------------------------------
def main():
    '''
    Starts here
    '''
    _get_args()
    _initialize()

    if Data.prnt:
        log.debug('Printing and returning')
        _print_configs()
        return

    cfg = _get_updated_config()
    _dump_config(cfg)
#---------------------------------------
if __name__ == '__main__':
    main()



'''
Module with utility functions to handle numpy arrays
'''

import math
import numpy

#-----------------------------------------------
def _check_ftimes(ftimes : float) -> None:
    '''
    Check if floating scale factor makes sense
    '''
    if not isinstance(ftimes, float):
        raise TypeError(f'Scaling factor is not a float, but: {ftimes}')

    if ftimes <= 1.0:
        raise ValueError(f'Scaling factor needs to be larger than 1.0, found: {ftimes}')
#-----------------------------------------------
def repeat_arr(arr_val : numpy.ndarray, ftimes : float) -> numpy.ndarray:
    '''
    Will repeat elements in an array a non integer number of times.

    arr_val: 1D array of objects
    ftimes (float): Number of times to repeat it.
    '''

    _check_ftimes(ftimes)

    a = math.floor(ftimes)
    b = math.ceil(ftimes)
    if numpy.isclose(a, b):
        return numpy.repeat(arr_val, a)

    # Will split randomly data in arr_val, such that one set will get increased
    # by floor(ftimes) and the other by ceiling(ftimes)

    # Get probability that given element belongs to dataset weighted by "a"
    p         = b - ftimes
    size_t    = len(arr_val)
    size_1    = int(p * size_t)

    # Find subset to weight by "a"
    arr_ind_1 = numpy.random.choice(size_t, size=size_1, replace=False)
    arr_val_1 = arr_val[arr_ind_1]

    # Find subset to weight by "b"
    arr_ind_2 = numpy.setdiff1d(numpy.arange(size_t), arr_ind_1)
    arr_val_2 = arr_val[arr_ind_2]

    # Repeat them an integer number of times
    arr_val_1 = numpy.repeat(arr_val_1, a)
    arr_val_2 = numpy.repeat(arr_val_2, b)

    return numpy.concatenate([arr_val_1, arr_val_2])
#-----------------------------------------------



'''
Module with utilities needed to manipulate ROOT files
'''
from ROOT import TTree, TDirectoryFile

def get_trees_from_file(ifile : TDirectoryFile) -> dict[str,TTree]:
    '''
    Picks up a TFile object
    Returns a dictionary of trees, with the tree location as the key
    Can search recursively within directories
    '''
    if not ifile.InheritsFrom('TDirectoryFile'):
        str_type = str(type(ifile))
        raise ValueError(f'Unrecognized object type {str_type}')

    dir_name = ifile.GetName()

    d_tree={}
    l_key =ifile.GetListOfKeys()

    for key in l_key:
        obj=key.ReadObj()
        if   obj.InheritsFrom('TDirectoryFile'):
            d_tmp = get_trees_from_file(obj)
            d_tree.update(d_tmp)
        elif obj.InheritsFrom('TTree'):
            obj_name = obj.GetName()
            key      = f'{dir_name}/{obj_name}'

            d_tree[key] = obj
        else:
            continue

    return d_tree



'''
Module with AtrMgr class
'''

import os

from ROOT import RDataFrame

import dmu.generic.utilities as gut
from dmu.logging.log_store import LogStore

#TODO:Skip attributes that start with Take< in a betterway
log = LogStore.add_logger('dmu:rdataframe:atr_mgr')
#------------------------
class AtrMgr:
    '''
    Class intended to store attributes of ROOT dataframes and attach them back after a Filtering, definition, redefinition, etc operation 
    These operations create new dataframes and therefore drop attributes.
    '''
    #------------------------
    def __init__(self, rdf : RDataFrame):
        self.d_in_atr = {}

        self._store_atr(rdf)
    #------------------------
    def _store_atr(self, rdf : RDataFrame):
        self.d_in_atr = self._get_atr(rdf)
    #------------------------
    def _skip_attr(self, name : str) -> bool:
        if name.startswith('__') and name.endswith('__'):
            return True

        return False
    #------------------------
    def _get_atr(self, rdf : RDataFrame):
        l_atr = dir(rdf)
        d_atr = {}
        for atr in l_atr:
            if self._skip_attr(atr):
                continue

            val = getattr(rdf, atr)
            d_atr[atr] = val

        return d_atr
    #------------------------
    def add_atr(self, rdf : RDataFrame) -> RDataFrame:
        '''
        Takes dataframe and adds back the attributes
        '''
        d_ou_atr = self._get_atr(rdf)

        key_in_atr = set(self.d_in_atr.keys())
        key_ou_atr = set(     d_ou_atr.keys())

        key_to_add = key_in_atr.difference(key_ou_atr)

        for key in key_to_add:
            val = self.d_in_atr[key]
            if key.startswith('Take<') and key.endswith('>'):
                continue

            log.info(f'Adding attribute {key}')
            setattr(rdf, key, val)

        return rdf
    #------------------------
    def to_json(self, json_path : str) -> None:
        '''
        Saves the attributes inside current instance to JSON. Takes JSON path as argument
        '''
        json_dir = os.path.dirname(json_path)
        os.makedirs(json_dir, exist_ok=True)

        t_type   = (list, str, int, float)
        d_fl_atr = { key : val for key, val in self.d_in_atr.items() if isinstance(val, t_type) and isinstance(key, t_type) }

        gut.dump_json(d_fl_atr, json_path)
#------------------------



'''
Module with utility functions related to the dmu.stats project
'''
import os
import re
from typing import Union
import zfit

from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:stats:utilities')
#-------------------------------------------------------
#Zfit/print_pdf
#-------------------------------------------------------
def _get_const(par : zfit.Parameter, d_const : Union[None, dict[str, list[float]]]) -> str:
    '''
    Takes zfit parameter and dictionary of constraints
    Returns a formatted string with the value of the constraint on that parameter
    '''
    if d_const is None or par.name not in d_const:
        return 'none'

    obj = d_const[par.name]
    if isinstance(obj, (list, tuple)):
        [mu, sg] = obj
        val      = f'{mu:.3e}; {sg:.3e}'
    else:
        val      = str(obj)

    return val
#-------------------------------------------------------
def _blind_vars(s_par : set, l_blind : Union[list[str], None] = None) -> set[zfit.Parameter]:
    '''
    Takes set of zfit parameters and list of parameter names to blind
    returns set of zfit parameters that should be blinded
    '''
    if l_blind is None:
        return s_par

    rgx_ors = '|'.join(l_blind)
    regex   = f'({rgx_ors})'

    s_par_blind = { par for par in s_par if not re.match(regex, par.name) }

    return s_par_blind
#-------------------------------------------------------
def _get_pars(
        pdf : zfit.pdf.BasePDF,
        blind : Union[None, list[str]]) -> tuple[list, list]:

    s_par_flt = pdf.get_params(floating= True)
    s_par_fix = pdf.get_params(floating=False)

    s_par_flt = _blind_vars(s_par_flt, l_blind=blind)
    s_par_fix = _blind_vars(s_par_fix, l_blind=blind)

    l_par_flt = list(s_par_flt)
    l_par_fix = list(s_par_fix)

    l_par_flt = sorted(l_par_flt, key=lambda par: par.name)
    l_par_fix = sorted(l_par_fix, key=lambda par: par.name)

    return l_par_flt, l_par_fix
#-------------------------------------------------------
def _get_messages(
        pdf       : zfit.pdf.BasePDF,
        l_par_flt : list,
        l_par_fix : list,
        d_const   : Union[None, dict[str,list[float]]] = None) -> list[str]:

    str_space = str(pdf.space)

    l_msg=[]
    l_msg.append('-' * 20)
    l_msg.append(f'PDF: {pdf.name}')
    l_msg.append(f'OBS: {str_space}')
    l_msg.append(f'{"Name":<50}{"Value":>15}{"Low":>15}{"High":>15}{"Floating":>5}{"Constraint":>25}')
    l_msg.append('-' * 20)
    for par in l_par_flt:
        value = par.value().numpy()
        low   = par.lower
        hig   = par.upper
        const = _get_const(par, d_const)
        l_msg.append(f'{par.name:<50}{value:>15.3e}{low:>15.3e}{hig:>15.3e}{par.floating:>5}{const:>25}')

    l_msg.append('')

    for par in l_par_fix:
        value = par.value().numpy()
        low   = par.lower
        hig   = par.upper
        const = _get_const(par, d_const)
        l_msg.append(f'{par.name:<50}{value:>15.3e}{low:>15.3e}{hig:>15.3e}{par.floating:>5}{const:>25}')

    return l_msg
#-------------------------------------------------------
def print_pdf(
        pdf      : zfit.pdf.BasePDF,
        d_const  : Union[None, dict[str,list[float]]] = None,
        txt_path : Union[str,None]                    = None,
        level    : int                                = 20,
        blind    : Union[None, list[str]]             = None):
    '''
    Function used to print zfit PDFs

    Parameters
    -------------------
    pdf (zfit.PDF): PDF
    d_const (dict): Optional dictionary mapping {par_name : [mu, sg]}
    txt_path (str): Optionally, dump output to text in this path
    level (str)   : Optionally set the level at which the printing happens in screen, default info
    blind (list)  : List of regular expressions matching variable names to blind in printout
    '''
    l_par_flt, l_par_fix = _get_pars(pdf, blind)
    l_msg                = _get_messages(pdf, l_par_flt, l_par_fix, d_const)

    if txt_path is not None:
        log.debug(f'Saving to: {txt_path}')
        message  = '\n'.join(l_msg)
        dir_path = os.path.dirname(txt_path)
        os.makedirs(dir_path, exist_ok=True)
        with open(txt_path, 'w', encoding='utf-8') as ofile:
            ofile.write(message)

        return

    for msg in l_msg:
        if   level == 20:
            log.info(msg)
        elif level == 30:
            log.debug(msg)
        else:
            raise ValueError(f'Invalid level: {level}')
#-------------------------------------------------------



'''
Module containing utility functions to be used with ROOT dataframes
'''

import re
from dataclasses import dataclass

import awkward as ak
import numpy

from ROOT import RDataFrame

from dmu.logging.log_store import LogStore

log = LogStore.add_logger('dmu:rdataframe:utilities')

# ---------------------------------------------------------------------
@dataclass
class Data:
    '''
    Class meant to store data that is shared
    '''
    l_good_type = [int, numpy.bool_, numpy.int32, numpy.uint32, numpy.int64, numpy.uint64, numpy.float32, numpy.float64]
    d_cast_type = {'bool': numpy.int32}
# ---------------------------------------------------------------------
def add_column(rdf : RDataFrame, arr_val : numpy.ndarray | None, name : str, d_opt : dict | None = None):
    '''
    Will take a dataframe, an array of numbers and a string
    Will add the array as a colunm to the dataframe

    d_opt (dict) : Used to configure adding columns
         exclude_re : Regex with patter of column names that we won't pick
    '''

    d_opt = {} if d_opt is None else d_opt
    if arr_val is None:
        raise ValueError('Array of values not introduced')

    if 'exclude_re' not in d_opt:
        d_opt['exclude_re'] = None

    v_col_org = rdf.GetColumnNames()
    l_col_org = [name.c_str() for name in v_col_org ]
    l_col     = []

    tmva_rgx  = r'tmva_\d+_\d+'

    for col in l_col_org:
        user_rgx = d_opt['exclude_re']
        if user_rgx is not None and re.match(user_rgx, col):
            log.debug(f'Dropping: {col}')
            continue

        if                          re.match(tmva_rgx, col):
            log.debug(f'Dropping: {col}')
            continue

        log.debug(f'Picking: {col}')
        l_col.append(col)

    data  = ak.from_rdataframe(rdf, columns=l_col)
    d_data= { col : data[col] for col in l_col }

    if arr_val.dtype == 'object':
        arr_val = arr_val.astype(float)

    d_data[name] = arr_val

    rdf = ak.to_rdataframe(d_data)

    return rdf
# ---------------------------------------------------------------------



'''
Script meant to do truth matching checks
'''
import os
import copy
import argparse

import yaml
import mplhep
import matplotlib.pyplot as plt

from ROOT import RDataFrame

from dmu.logging.log_store   import LogStore
from dmu.plotting.plotter_1d import Plotter1D as Plotter

log=LogStore.add_logger('dmu:physics:check_truth')
# ----------------------------------
def _set_logs() -> None:
    LogStore.set_level('dmu:plotting:Plotter'  , 30)
    LogStore.set_level('dmu:plotting:Plotter1D', 30)
# ----------------------------------
def _get_args() -> argparse.Namespace:
    '''
    Parse args
    '''
    parser = argparse.ArgumentParser(description='Script used to carry out checks on truth matching mechanisms for MC')
    parser.add_argument('-c', '--conf' , type=str, help='Path to config file', required=True)
    args = parser.parse_args()

    return args
# ----------------------------------
def _get_config(args : argparse.Namespace) -> dict:
    path = args.conf
    if not os.path.isfile(path):
        raise FileNotFoundError(f'Cannot find {path}')

    with open(path, encoding='utf-8') as ifile:
        cfg = yaml.safe_load(ifile)

    return cfg
# ----------------------------------
def _get_rdf(file_path : str, tree_path : str) -> RDataFrame:
    log.debug(f'Picking inputs from: {file_path}/{tree_path}')
    rdf = RDataFrame(tree_path, file_path)

    nentries = rdf.Count().GetValue()
    log.debug(f'Found {nentries} entries')

    return rdf
# ----------------------------------
def _preprocess_rdf(rdf : RDataFrame, cfg : dict) -> RDataFrame:
    if 'max_entries' in cfg:
        max_entries = cfg['max_entries']
        rdf = rdf.Range(max_entries)

    return rdf
# ----------------------------------
def _check(cfg : dict) -> None:
    log.info(110 * '-')
    log.info(f'{"Sample":<20}{"Method":<20}{"Initial":<15}{"":<15}{"Final":<15}{"":15}{"Efficiency":<10}')
    log.info(110 * '-')

    for sample_name in cfg['samples']:
        file_path = cfg['samples'][sample_name]['file_path']
        tree_path = cfg['samples'][sample_name]['tree_path']
        rdf = _get_rdf(file_path, tree_path)
        rdf = _preprocess_rdf(rdf, cfg)

        d_cut_true = {}
        d_cut_fake = {}
        for method, cut in cfg['samples'][sample_name]['methods'].items():
            _check_kind(rdf, sample_name, method, cut)

            d_cut_true[method] = cut
            d_cut_fake[method] = f'({cut}) == 0'
        log.info('')

        _plot_distributions(cfg, sample_name, rdf, d_cut_true, kind='matched')
        _plot_distributions(cfg, sample_name, rdf, d_cut_fake, kind='anti_matched')
# ----------------------------------
def _plot_distributions(cfg : dict, sample_name : str, rdf : RDataFrame, d_cut : dict[str,str], kind : str) -> None:
    cfg     = copy.deepcopy(cfg)
    cfg_plt = cfg['samples'][sample_name]['plot']
    cfg_plt = _add_suffix(cfg_plt, sample_name, kind)
    d_rdf   = { method : rdf.Filter(cut) for method, cut in d_cut.items() }

    ptr=Plotter(d_rdf=d_rdf, cfg=cfg_plt)
    ptr.run()
# ----------------------------------
def _add_suffix(cfg : dict, sample_name : str, kind : str) -> dict:
    d_var = cfg['plots']
    for var in d_var:
        d_var[var]['name']  = f'{var}_{kind}'
        d_var[var]['title'] = f'{sample_name}; {kind}'

    cfg['plots'] = d_var

    return cfg
# ----------------------------------
def _check_kind(rdf : RDataFrame, sample : str, name : str, cut : str) -> RDataFrame:
    nini = rdf.Count().GetValue()
    rdf  = rdf.Filter(cut, name)
    nfnl = rdf.Count().GetValue()
    eff  = nfnl / nini * 100

    log.info(f'{sample:<20}{name:<20}{nini:<15}{"":<15}{nfnl:<15}{"-->":15}{eff:10.2f}')
# ----------------------------------
def main():
    '''
    Script starts here
    '''
    _set_logs()
    args = _get_args()
    cfg  = _get_config(args)
    plt.style.use(mplhep.style.LHCb2)

    _check(cfg)
# ----------------------------------
if __name__ == '__main__':
    main()



'''
Module containing Plotter2D class
'''
from typing import Union

import hist
import numpy
import mplhep
import matplotlib.pyplot as plt

from hist                  import Hist
from ROOT                  import RDataFrame
from dmu.logging.log_store import LogStore
from dmu.plotting.plotter  import Plotter

log = LogStore.add_logger('dmu:plotting:Plotter2D')
# --------------------------------------------
class Plotter2D(Plotter):
    '''
    Class used to plot columns in ROOT dataframes
    '''
    # --------------------------------------------
    def __init__(self, rdf=None, cfg=None):
        '''
        Parameters:

        d_rdf (dict): Dictionary mapping the kind of sample with the ROOT dataframe
        cfg   (dict): Dictionary with configuration, e.g. binning, ranges, etc
        '''

        if not isinstance(cfg, dict):
            raise ValueError('Config dictionary not passed')

        self._rdf   : RDataFrame = rdf
        self._d_cfg : dict       = cfg

        self._wgt : numpy.ndarray
    # --------------------------------------------
    def _get_axis(self, var : str):
        [minx, maxx, nbins] = self._d_cfg['axes'][var]['binning']
        label               = self._d_cfg['axes'][var][  'label']

        axis = hist.axis.Regular(nbins, minx, maxx, name=label, label=label)

        return axis
    # --------------------------------------------
    def _get_data(self, varx : str, vary : str) -> tuple[numpy.ndarray, numpy.ndarray]:
        d_data = self._rdf.AsNumpy([varx, vary])
        arr_x  = d_data[varx]
        arr_y  = d_data[vary]

        return arr_x, arr_y
    # --------------------------------------------
    def _get_dataset_weights(self, wgt_name : Union[str, None]) -> Union[numpy.ndarray, None]:
        if wgt_name is None:
            log.debug('Skipping weights')
            return None

        log.debug(f'Adding weights form column {wgt_name}')
        arr_wgt  = self._rdf.AsNumpy([wgt_name])[wgt_name]

        return arr_wgt
    # --------------------------------------------
    def _plot_vars(self, varx : str, vary : str, wgt_name : str) -> None:
        log.info(f'Plotting {varx} vs {vary} with weights {wgt_name}')

        ax_x         = self._get_axis(varx)
        ax_y         = self._get_axis(vary)
        arr_x, arr_y = self._get_data(varx, vary)

        arr_w = self._get_dataset_weights(wgt_name)
        hst   = Hist(ax_x, ax_y)
        hst.fill(arr_x, arr_y, weight=arr_w)

        mplhep.hist2dplot(hst)
    # --------------------------------------------
    def run(self):
        '''
        Will run plotting
        '''

        fig_size = self._get_fig_size()
        for [varx, vary, wgt_name, plot_name] in self._d_cfg['plots_2d']:
            plt.figure(plot_name, figsize=fig_size)
            self._plot_vars(varx, vary, wgt_name)
            self._save_plot(plot_name)
# --------------------------------------------



'''
Module holding cv_classifier class
'''

from sklearn.ensemble        import GradientBoostingClassifier

from dmu.logging.log_store import LogStore
import dmu.ml.utilities    as ut

log = LogStore.add_logger('dmu:ml:CVClassifier')

# ---------------------------------------
class CVSameData(Exception):
    '''
    Will be raised if a model is been evaluated with a dataset such that at least one of the
    samples was also used for the training
    '''
# ---------------------------------------
class CVClassifier(GradientBoostingClassifier):
    '''
    Derived class meant to implement features needed for cross-validation
    '''
    # pylint: disable = too-many-ancestors, abstract-method
    # ----------------------------------
    def __init__(self, cfg : dict | None = None):
        '''
        cfg (dict) : Dictionary with configuration, specially the hyperparameters set in the `hyper` field
        '''
        if cfg is None:
            raise ValueError('No configuration was passed')

        self._cfg = cfg

        d_hyp = self._cfg['training']['hyper']
        super().__init__(**d_hyp)

        self._s_hash    = set()
        self._data      = {}
        self._l_ft_name = None
    # ----------------------------------
    @property
    def features(self):
        '''
        Returns list of feature names used in training dataset
        '''
        return self._l_ft_name
    # ----------------------------------
    @property
    def hashes(self):
        '''
        Will return set with hashes of training data
        '''
        return self._s_hash
    # ----------------------------------
    @property
    def cfg(self):
        '''
        Will return dictionary with configuration
        '''

        return self._cfg
    # ----------------------------------
    def __str__(self):
        nhash = len(self._s_hash)

        msg = 40 * '-' + '\n'
        msg+= f'{"Attribute":<20}{"Value":<20}\n'
        msg+= 40 * '-' + '\n'
        msg += f'{"Hashes":<20}{nhash:<20}\n'
        msg+= 40 * '-'

        return msg
    # ----------------------------------
    def fit(self, *args, **kwargs):
        '''
        Runs the training of the model
        '''
        log.debug('Fitting')

        df_ft           = args[0]
        self._l_ft_name = list(df_ft.columns)

        self._s_hash = ut.get_hashes(df_ft)
        log.debug(f'Saving {len(self._s_hash)} hashes')

        super().fit(*args, **kwargs)

        return self
    # ----------------------------------
    def _check_hashes(self, df_ft):
        '''
        Will check that the hashes of the passed features do not intersect with the
        hashes of the features used for the training.
        Else it will raise CVSameData exception
        '''

        if len(self._s_hash) == 0:
            raise ValueError('Found no hashes in model')

        s_hash  = ut.get_hashes(df_ft)
        s_inter = self._s_hash.intersection(s_hash)

        nh1 = len(self._s_hash)
        nh2 = len(      s_hash)
        nh3 = len(s_inter)

        if nh3 > 0:
            raise CVSameData(f'Found non empty intersection of size: {nh1} ^ {nh2} = {nh3}')
    # ----------------------------------
    def predict_proba(self, X, on_training_ok=False):
        '''
        Takes pandas dataframe with features
        Will first check hashes to make sure none of the events/samples
        used for the training of this model are in the prediction

        on_training_ok (bool): True if the dataset is expected to contain samples used for training, default is False
        '''
        if not on_training_ok:
            self._check_hashes(X)

        return super().predict_proba(X)
# ---------------------------------------
